from space_tracer.main import replace_input, TraceRunner


def test_source_width_positive():
    code = """\
i = 1 + 1
"""
    expected_report = """\
i = 1 +  | i = 2"""

    with replace_input(code):
        report = TraceRunner().trace_command(['space_tracer',
                                              '--source_width', '8',
                                              '--traced_file', 'foo.py'])

    assert report == expected_report


def test_source_width_negative():
    code = """\
i = 1 + 1
"""
    expected_report = """\
i = 1 + | i = 2"""

    with replace_input(code):
        report = TraceRunner().trace_command(['space_tracer',
                                              '--source_width', '-2',
                                              '--traced_file', 'foo.py'])

    assert report == expected_report


def test_source_indent():
    code = """\
i = 1 + 1
"""
    expected_report = """\
    i = 1 + 1 | i = 2"""

    with replace_input(code):
        report = TraceRunner().trace_command(['space_tracer',
                                              '--source_indent', '4',
                                              '--traced_file', 'foo.py'])

    assert report == expected_report


def test_source_indent_small():
    code = """\
i = 1 + 1
"""
    expected_report = """\
  i = 1 + 1 | i = 2"""

    with replace_input(code):
        report = TraceRunner().trace_command(['space_tracer',
                                              '--source_indent', '2',
                                              '--traced_file', 'foo.py'])

    assert report == expected_report


def test_source_indent_negative():
    code = """\
i = 1 + 1
"""
    expected_report = """\
= 1 + 1 | i = 2"""

    with replace_input(code):
        report = TraceRunner().trace_command(['space_tracer',
                                              '--source_indent', '-2',
                                              '--traced_file', 'foo.py'])

    assert report == expected_report


def test_trace_width():
    code = """\
i = 1 + 1
"""
    expected_report = """\
i = 1 + 1 | i ="""

    with replace_input(code):
        report = TraceRunner().trace_command(['space_tracer',
                                              '--trace_width', '15',
                                              '--traced_file', 'foo.py'])

    assert report == expected_report


def test_trace_width_negative():
    code = """\
i = 1 + 1
s = 'a' * 10
"""
    expected_report = """\
i = 1 + 1    | i = 2
s = 'a' * 10 | s = 'aaaaaa"""

    with replace_input(code):
        report = TraceRunner().trace_command(['space_tracer',
                                              '--trace_width', '-5',
                                              '--traced_file', 'foo.py'])

    assert report == expected_report


def test_trace_width_without_source():
    code = """\
i = 1 + 1
s = 'a' * 10
"""
    expected_report = """\
i = 2
s = 'aaaaaa"""

    with replace_input(code):
        report = TraceRunner().trace_command(['space_tracer',
                                              '--source_width', '0',
                                              '--trace_width', '-5',
                                              '--traced_file', 'foo.py'])

    assert report == expected_report


def test_trace_offset():
    code = """\
i = 1 + 1
s = 'a' * 10
"""
    expected_report = """\
i = 1 + 1    |  2
s = 'a' * 10 |  'aaaaaaaaaa'"""

    with replace_input(code):
        report = TraceRunner().trace_command(['space_tracer',
                                              '--trace_offset', '3',
                                              '--traced_file', 'foo.py'])

    assert report == expected_report



import sys

import pytest

from space_tracer.main import TraceRunner


def test_function():
    code = """\
def foo(x):
    a = x
    b = x + 1
    return b

m = 2
n = foo(m)
"""
    expected_report = """\
x = 2
a = 2
b = 3
return 3

m = 2
n = 3"""
    tracer = TraceRunner()

    report = tracer.trace_code(code)

    assert report == expected_report


def test_empty_return():
    code = """\
def foo(x):
    return

n = foo(10)
"""
    expected_report = """\
x = 10


n = None"""
    tracer = TraceRunner()

    report = tracer.trace_code(code)

    assert report == expected_report


def test_empty_function():
    code = '''\
def foo(x):
    """ No function body, just a docstring. """

n = foo(10)
'''
    expected_report = """\
x = 10


n = None"""
    tracer = TraceRunner()

    report = tracer.trace_code(code)

    assert report == expected_report


def test_return_subscript():
    code = """\
def foo(x):
    a = [x, 3]
    return a[1]

n = foo(2)
"""
    expected_report = """\
x = 2
a = [2, 3]
return 3

n = 3"""
    tracer = TraceRunner()

    report = tracer.trace_code(code)

    assert report == expected_report


@pytest.mark.skipif(sys.version_info < (3, 8),
                    reason='Positional-only params not supported before 3.8.')
def test_positional_parameters():
    code = """\
def foo(a, /, b):
    print(a, b)

foo(1, 2)
"""
    expected_report = """\
a = 1 | b = 2
print('1 2')

"""
    tracer = TraceRunner()

    report = tracer.trace_code(code)

    assert report == expected_report


def test_keyword_only_parameters():
    code = """\
def foo(a, *, b):
    print(a, b)

foo(1, b=2)
"""
    expected_report = """\
a = 1 | b = 2
print('1 2')

"""
    tracer = TraceRunner()

    report = tracer.trace_code(code)

    assert report == expected_report


def test_method_call():
    code = """\
a = [2, 1]
a.sort()
a.sort() # second call makes no change, nothing printed
"""
    expected_report = """\
a = [2, 1]
a = [1, 2]
"""

    report = TraceRunner().trace_code(code)

    assert report == expected_report


def test_nested_method_call():
    code = """\
class Foo(object):
    pass

f = Foo()
f.items = []
f.items.append(2)
"""
    expected_report = """\




f.items = []
f.items = [2]"""

    report = TraceRunner().trace_code(code)

    assert report == expected_report


def test_method_call_output_param():
    code = """\
from heapq import heappush
l = []
heappush(l, 5)
"""
    expected_report = """\

l = []
l = [5]"""

    report = TraceRunner().trace_code(code)

    assert report == expected_report


def test_method_call_output_param_keyword():
    code = """\
def grow(l):
    l.append(0)
a = []
grow(l=a)
"""
    expected_report = """\
l = []
l = [0]
a = []
a = [0]"""

    report = TraceRunner().trace_code(code)

    assert report == expected_report


def test_incomplete_iterator():
    code = """\
def gen(n):
    state = 'Starting'
    try:
        for i in range(n):
            yield i
    finally:
        state = 'Done'


g = gen(999)
x = next(g)
"""

    expected_report = """\
n = 999
state = 'Starting'

i = 0
yield 0 GeneratorExit

state = 'Done'



x = 0"""
    tracer = TraceRunner()

    report = tracer.trace_code(code)

    assert report == expected_report


def test_incomplete_iterator_prints():
    code = """\
def gen(n):
    print('Starting')
    try:
        for i in range(n):
            yield i
    finally:
        print('Done')


g = gen(999)
x = next(g)
"""

    expected_report = """\
n = 999
print('Starting')

i = 0
yield 0 GeneratorExit

print('Done')



x = 0"""
    tracer = TraceRunner()

    report = tracer.trace_code(code)

    assert report == expected_report



import pytest

from space_tracer.canvas import Canvas
from space_tracer.mock_turtle import MockTurtle
import turtle


@pytest.fixture
def patched_turtle():
    MockTurtle.monkey_patch()
    yield
    MockTurtle.remove_monkey_patch()


def test_forward(patched_turtle):
    expected_report = """\
create_line
    0
    0
    100
    0
    fill='black'
    pensize=1
"""

    t = MockTurtle()
    t.fd(100)
    report = t.report

    assert report == expected_report.splitlines()


def test_right(patched_turtle):
    expected_report = """\
create_line
    0
    0
    0
    100
    fill='black'
    pensize=1
"""

    t = MockTurtle()
    t.right(90)
    t.fd(100)
    report = t.report

    assert report == expected_report.splitlines()


def test_penup(patched_turtle):
    expected_report = """\
create_line
    0
    0
    100
    0
    fill='black'
    pensize=1
create_line
    150
    0
    350
    0
    fill='#000000'
    pensize=1
"""

    t = MockTurtle()
    t.fd(100)
    t.penup()
    t.fd(50)
    t.pendown()
    t.fd(200)
    report = t.report

    assert report == expected_report.splitlines()


def test_dot(patched_turtle):
    t2 = MockTurtle()
    t2.up()
    t2.goto(0, -2.5)
    t2.begin_fill()
    t2.circle(2.5)
    t2.end_fill()
    expected_report = t2.report
    MockTurtle._screen = None

    t = MockTurtle()
    t.dot()
    report = t.report

    assert report == expected_report


def test_bounds():
    expected_width = 800
    expected_height = 600

    t = MockTurtle(canvas=Canvas(expected_width, expected_height))
    width = t.window_width()
    height = t.window_height()

    assert width == expected_width
    assert height == expected_height


def test_bounds_after_monkey_patch():
    expected_width = 300
    expected_height = 200

    MockTurtle.monkey_patch(canvas=Canvas(expected_width, expected_height))
    try:
        width = turtle.window_width()
        height = turtle.window_height()
    finally:
        MockTurtle.remove_monkey_patch()

    assert width == expected_width
    assert height == expected_height


def test_offset():
    MockTurtle.remove_monkey_patch()
    expected_report = """\
create_line
    400
    300
    500
    300
    fill='black'
    pensize=1
"""

    t = MockTurtle(canvas=Canvas(800, 600))
    t.fd(100)
    report = t.report

    assert report == expected_report.splitlines()


def test_scale(patched_turtle):
    expected_report = """\
create_line
    0
    0
    100
    0
    fill='black'
    pensize=1
create_line
    100
    0
    100
    150
    fill='black'
    pensize=1
"""

    t = MockTurtle(canvas=Canvas())
    t.screen.xscale = 100.0
    t.screen.yscale = 50
    t.fd(1)
    t.right(90)
    t.fd(3)
    report = t.report

    assert report == expected_report.splitlines()


def test_offset_with_scale():
    """ The offset is applied BEFORE the scale. """

    MockTurtle.remove_monkey_patch()
    expected_report = """\
create_line
    400
    300
    500
    300
    fill='black'
    pensize=1
"""

    t = MockTurtle(canvas=Canvas(800, 600))
    t.screen.xscale = 100
    t.fd(1)
    report = t.report

    assert report == expected_report.splitlines()


def test_repr():
    expected_text = "MockTurtle(100, 0, 10)"

    t = MockTurtle(25, 0, -7)
    t.left(7)
    t.fd(75)
    t.left(10)
    text = repr(t)

    assert text == expected_text


def test_write(patched_turtle):
    expected_report = """\
create_line
    0
    0
    100
    0
    fill='black'
    pensize=1
create_text
    100
    0
    anchor='sw'
    fill='black'
    font=('Arial', 8, 'normal')
    text='Bob'
"""

    t = MockTurtle()
    t.fd(100)
    t.write('Bob')
    report = t.report

    assert report == expected_report.splitlines()


def test_write_center(patched_turtle):
    expected_report = """\
create_line
    0
    0
    100
    0
    fill='black'
    pensize=1
create_text
    100
    0
    anchor='s'
    fill='black'
    font=('Arial', 8, 'normal')
    text='Bob'
"""

    t = MockTurtle()
    t.fd(100)
    t.write('Bob', align='center')
    report = t.report

    assert report == expected_report.splitlines()


def test_write_right(patched_turtle):
    expected_report = """\
create_line
    0
    0
    100
    0
    fill='black'
    pensize=1
create_text
    100
    0
    anchor='se'
    fill='black'
    font=('Arial', 8, 'normal')
    text='Bob'
"""

    t = MockTurtle()
    t.fd(100)
    t.write('Bob', align='right')
    report = t.report

    assert report == expected_report.splitlines()


def test_write_font(patched_turtle):
    expected_report = """\
create_line
    0
    0
    100
    0
    fill='black'
    pensize=1
create_text
    100
    0
    anchor='sw'
    fill='black'
    font=('Courier', 14, 'bold')
    text='Bob'
"""

    t = MockTurtle()
    t.fd(100)
    t.write('Bob', font=('Courier', 14, 'bold'))
    report = t.report

    assert report == expected_report.splitlines()


def test_write_partial_font(patched_turtle):
    expected_report = """\
create_line
    0
    0
    100
    0
    fill='black'
    pensize=1
create_text
    100
    0
    anchor='sw'
    fill='black'
    font=('Courier', 8, 'normal')
    text='Bob'
"""

    t = MockTurtle()
    t.fd(100)
    t.write('Bob', font=('Courier',))
    report = t.report

    assert report == expected_report.splitlines()


def test_write_font_string(patched_turtle):
    expected_report = """\
create_line
    0
    0
    100
    0
    fill='black'
    pensize=1
create_text
    100
    0
    anchor='sw'
    fill='black'
    font=('Courier', 8, 'normal')
    text='Bob'
"""

    t = MockTurtle()
    t.fd(100)
    t.write('Bob', font='Courier')
    report = t.report

    assert report == expected_report.splitlines()


def test_write_font_list(patched_turtle):
    expected_report = """\
create_line
    0
    0
    100
    0
    fill='black'
    pensize=1
create_text
    100
    0
    anchor='sw'
    fill='black'
    font=('Courier', 14, 'italic')
    text='Bob'
"""

    t = MockTurtle()
    t.fd(100)
    t.write('Bob', font=['Courier', 14, 'italic'])
    report = t.report

    assert report == expected_report.splitlines()


def test_write_font_none(patched_turtle):
    expected_report = """\
create_line
    0
    0
    100
    0
    fill='black'
    pensize=1
create_text
    100
    0
    anchor='sw'
    fill='black'
    font=('Arial', 8, 'normal')
    text='Bob'
"""

    t = MockTurtle()
    t.fd(100)
    t.write('Bob', font=None)
    report = t.report

    assert report == expected_report.splitlines()


def test_write_move():
    """ Not supported yet """
    t = MockTurtle()
    with pytest.raises(NotImplementedError):
        t.write('Bob', move=True)


def test_write_bad_size():
    t = MockTurtle()
    with pytest.raises(ValueError):
        t.write('Bob', font=('Arial', 'eight', 'normal'))


def test_color(patched_turtle):
    expected_report = """\
create_line
    0
    0
    100
    0
    fill='#ff0080'
    pensize=1"""

    t = MockTurtle()
    t.color(1.0, 0.0, 0.5)
    t.fd(100)
    report = t.report

    assert report == expected_report.splitlines()


def test_color_name(patched_turtle):
    expected_report = """\
create_line
    0
    0
    100
    0
    fill='#0000ff'
    pensize=1"""

    t = MockTurtle()
    t.color('blue')
    t.fd(100)
    report = t.report

    assert report == expected_report.splitlines()


def test_color_bad(patched_turtle):
    expected_report = """\
create_line
    0
    0
    100
    0
    fill='#000000'
    pensize=1"""

    t = MockTurtle()
    # noinspection PyTypeChecker
    t.color((1.0, 0.0))  # Only two numbers, fails to black.
    t.fd(100)
    report = t.report

    assert report == expected_report.splitlines()


def test_color_bad_range(patched_turtle):
    expected_report = """\
create_line
    0
    0
    100
    0
    fill='#000000'
    pensize=1"""

    t = MockTurtle()
    t.color(1.0, 0.0, 1.5)  # Over 1.0 not allowed, fails to black.
    t.fd(100)
    report = t.report

    assert report == expected_report.splitlines()


def test_pen_dict(patched_turtle):
    expected_report = """\
create_line
    0
    0
    100
    0
    fill='#0000ff'
    pensize=1"""

    t = MockTurtle()
    t.pen(pencolor=(0, 0, 1.0))
    t.fd(100)
    report = t.report

    assert report == expected_report.splitlines()


def test_get_color_names():
    t = MockTurtle()
    t.color('blue')
    color = t.color()

    # Before Python 3.6, dicts were not ordered, and either name was possible.
    assert color in (('blue', 'blue'), ('blue1', 'blue1'))


def test_get_color_rgb():
    t = MockTurtle()
    expected_color = (1.0, 0.0, 0.5)
    t.color(expected_color)
    color = t.color()

    assert color == (expected_color, expected_color)


def test_get_default_color():
    t = MockTurtle()
    color = t.color()

    assert color == ('black', 'black')


def test_bgcolor(patched_turtle):
    expected_report = """\
bgcolor
    fill='#00ff00'
    outline=''
create_line
    0
    0
    100
    0
    fill='black'
    pensize=1"""

    t = MockTurtle()
    t.fd(100)
    color1 = t.screen.bgcolor()
    t.screen.bgcolor('green')
    color2 = t.screen.bgcolor()
    report = t.report

    assert report == expected_report.splitlines()
    assert color1 == 'white'
    assert color2 == 'green'


def test_fill(patched_turtle):
    expected_report = """\
create_polygon
    0
    0
    100
    0
    100
    100
    fill='#0000ff'
    outline=''
create_line
    0
    0
    100
    0
    fill='#ff0000'
    pensize=1
create_line
    100
    0
    100
    100
    fill='#ff0000'
    pensize=1"""

    t = MockTurtle()
    t.color('red', 'blue')
    t.begin_fill()
    for _ in range(2):
        t.fd(100)
        t.right(90)
    t.end_fill()
    report = t.report

    assert report == expected_report.splitlines()


def test_is_filling():
    t = MockTurtle()
    is_filling1 = t.fill()
    t.begin_fill()
    is_filling2 = t.fill()
    t.end_fill()
    is_filling3 = t.fill()

    assert not is_filling1
    assert is_filling2
    assert not is_filling3


def test_forgotten_end_fill(patched_turtle):
    expected_report = """\
create_line
    0
    0
    100
    0
    fill='#ff0000'
    pensize=1
create_line
    100
    0
    100
    100
    fill='#ff0000'
    pensize=1
"""

    t = MockTurtle()
    t.color('red', 'blue')
    t.begin_fill()
    for _ in range(2):
        t.fd(100)
        t.right(90)
    report = t.report

    assert report == expected_report.splitlines()


def test_stamp(patched_turtle):
    expected_report = """\
create_polygon
    0
    0
    -9
    -5
    -7
    0
    -9
    5
    0
    0
    fill='#000000'
    outline=''
create_line
    0
    0
    -9
    -5
    fill='#000000'
    pensize=1
create_line
    -9
    -5
    -7
    0
    fill='#000000'
    pensize=1
create_line
    -7
    0
    -9
    5
    fill='#000000'
    pensize=1
create_line
    -9
    5
    0
    0
    fill='#000000'
    pensize=1
"""

    t = MockTurtle()
    t.stamp()
    report = t.report

    assert report == expected_report.splitlines()


def test_forgotten_end_fill_with_stamp(patched_turtle):
    expected_report = """\
create_polygon
    0
    0
    -9
    -5
    -7
    0
    -9
    5
    0
    0
    fill='#000000'
    outline=''
create_line
    0
    0
    -9
    -5
    fill='#000000'
    pensize=1
create_line
    -9
    -5
    -7
    0
    fill='#000000'
    pensize=1
create_line
    -7
    0
    -9
    5
    fill='#000000'
    pensize=1
create_line
    -9
    5
    0
    0
    fill='#000000'
    pensize=1
create_line
    0
    0
    100
    0
    fill='#ff0000'
    pensize=1
create_line
    100
    0
    100
    100
    fill='#ff0000'
    pensize=1
"""

    t = MockTurtle()
    t.stamp()
    t.color('red', 'blue')
    t.begin_fill()
    for _ in range(2):
        t.fd(100)
        t.right(90)
    report = t.report

    assert report == expected_report.splitlines()


def test_monkey_patch_anonymous_turtle(patched_turtle):
    expected_report = """\
create_line
    0
    0
    100
    0
    fill='black'
    pensize=1
"""

    turtle.fd(100)
    report = MockTurtle.get_all_reports()

    assert report == expected_report.splitlines()


def test_monkey_patch_new_turtle(patched_turtle):
    expected_report = """\
create_line
    0
    0
    100
    0
    fill='black'
    pensize=1
"""

    t = turtle.Turtle()
    t.fd(100)
    report = MockTurtle.get_all_reports()

    assert report == expected_report.splitlines()


def test_monkey_patch_multiple_turtles(patched_turtle):
    expected_report = """\
create_line
    0
    0
    100
    0
    fill='black'
    pensize=1
create_line
    100
    0
    100
    100
    fill='black'
    pensize=1
create_line
    0
    0
    100
    0
    fill='black'
    pensize=1
create_line
    100
    0
    100
    -100
    fill='black'
    pensize=1
"""

    t1 = turtle.Turtle()
    t1.begin_fill()
    t1.fd(100)
    t1.right(90)
    t1.fd(100)
    t2 = turtle.Turtle()
    t2.begin_fill()
    t2.fd(100)
    t2.left(90)
    t2.fd(100)
    report = MockTurtle.get_all_reports()

    assert report == expected_report.splitlines()


def test_bad_attribute():
    t = MockTurtle()

    assert not hasattr(t, 'bogus')


def test_screen_methods_exist():
    """ Test that a couple of methods exist, but don't do anything. """
    t = MockTurtle()

    t.screen.tracer()
    t.screen.update()



from unittest import TestCase

from example_source import bar


class Bucket(object):
    def __init__(self):
        self.contents = []

    def add(self, item):
        self.contents.append(item)


class BarTest(TestCase):
    def test_bar(self):
        bucket = Bucket()

        bar(bucket)

        self.assertEqual(['bar'], bucket.contents)



import doctest
import os
import re
import sys

import pytest
from unittest.mock import DEFAULT, patch

from space_tracer import main
from space_tracer.main import replace_input, TraceRunner
from test_report_builder import ReportTestCase

EXAMPLE_DRIVER_PATH = os.path.join(os.path.dirname(__file__),
                                   'example_driver.py')
EXAMPLE_SOURCE_PATH = os.path.join(os.path.dirname(__file__),
                                   'example_source.py')
EXAMPLE_PRINTING_PATH = os.path.join(os.path.dirname(__file__),
                                     'example_printing.py')
EXAMPLE_LIB_PATH = os.path.join(os.path.dirname(__file__),
                                'example_package',
                                'lib_in_package.py')
EXAMPLE_PATCHING_DRIVER_PATH = os.path.join(os.path.dirname(__file__),
                                            'example_patching_driver.py')
EXAMPLE_DRIVER_SYNTAX_ERROR_PATH = os.path.join(os.path.dirname(__file__),
                                                'example_driver_syntax_error.py')
EXAMPLE_PYCHARM_FAILURES_PATH = os.path.join(os.path.dirname(__file__),
                                             'example_pycharm_failures.py')
EXAMPLE_SILENT_DRIVER_PATH = os.path.join(os.path.dirname(__file__),
                                          'example_silent_driver.py')
patch.multiple = patch.multiple  # Avoids PyCharm warnings.


@pytest.fixture
def stdin():
    with patch('sys.stdin') as mocked:
        yield mocked


@pytest.fixture
def stdout(capsys):
    class MockIO(object):
        @staticmethod
        def getvalue():
            return capsys.readouterr().out
    yield MockIO()


@pytest.fixture
def stderr(capsys):
    class MockIO(object):
        @staticmethod
        def getvalue():
            return capsys.readouterr().err
    yield MockIO()


@pytest.fixture
def argv():
    mocked = []
    with patch('sys.argv', mocked):
        yield mocked


class CodeTracerMainTest(ReportTestCase):
    def setUp(self):
        super(CodeTracerMainTest, self).setUp()
        self.maxDiff = None
        for module_name in ('example_source',
                            'example_package',
                            'example_package.__main__',
                            'example_package.lib_in_package',
                            'example_driver'):
            if module_name in sys.modules:
                del sys.modules[module_name]

    @patch.multiple('sys', stdin=DEFAULT, stdout=DEFAULT, argv=[
        'dummy.py',
        '--source_width', '0',
        '--traced_file', 'foo.py'])
    def test_main(self, stdin, stdout):
        code = """\
i = 1
name = __name__
"""
        expected_report = """\
i = 1
name = '__main__' """
        stdin.read.return_value = code

        main()

        self.assertReportEqual(expected_report,
                               stdout.write.call_args_list[0][0][0])

    @patch.multiple('sys', stdin=DEFAULT, stdout=DEFAULT, argv=[
        'dummy.py',
        '--source_indent', '4',
        '--live',
        EXAMPLE_SOURCE_PATH])
    def test_source_file_arg(self, stdin, stdout):
        expected_report = """\
    def foo(x):                       | x = 3
        return x + 1                  | return 4
                                      |
                                      |
    def bar(bucket):                  |
        bucket.add('bar')             |
                                      |
                                      |
    if __name__ == '__live_coding__': |
        y = foo(3)                    | y = 4
"""
        stdin.read.return_value = ""

        main()

        report = stdout.write.call_args_list[0][0][0]
        self.assertReportEqual(expected_report, report)

    @patch.multiple('sys', stdin=DEFAULT, stdout=DEFAULT, argv=[
        'dummy.py',
        '--source_width', '0',
        '--traced_file', EXAMPLE_SOURCE_PATH,
        '--traced', 'example_source',
        EXAMPLE_DRIVER_PATH])
    def test_driver(self, stdin, stdout):
        source = """\
def foo(x):
    name = __name__
    return x + 1
"""
        expected_report = """\
x = 42
name = 'example_source'
return 43
"""
        stdin.read.return_value = source

        main()

        report = stdout.write.call_args_list[0][0][0]
        self.assertReportEqual(expected_report, report)

    @patch.multiple('sys', stdin=DEFAULT, stdout=DEFAULT, argv=[
        'dummy.py',
        '--source_width', '0',
        '--traced_file', EXAMPLE_SOURCE_PATH,
        '--traced', 'example_source',
        EXAMPLE_DRIVER_PATH,
        '99'])
    def test_driver_args(self, stdin, stdout):
        source = """\
import sys
def foo(x):
    return sys.argv[1:]
"""
        expected_report = """\

x = 42
return ['99']
"""
        stdin.read.return_value = source

        main()

        report = stdout.write.call_args_list[0][0][0]
        self.assertReportEqual(expected_report, report)

    @patch.multiple('sys', stdin=DEFAULT, stdout=DEFAULT, argv=[
        'dummy.py',
        '--source_width', '0',
        '--traced_file', EXAMPLE_SOURCE_PATH,
        '--traced', '__live_coding__',
        '--live',
        EXAMPLE_SOURCE_PATH,
        '99'])
    def test_args_no_driver(self, stdin, stdout):
        source = """\
import sys
x = sys.argv[1:]
"""
        expected_report = """\

x = ['99']
"""
        stdin.read.return_value = source

        main()

        report = stdout.write.call_args_list[0][0][0]
        self.assertReportEqual(expected_report, report)

    @patch.multiple('sys', stdin=DEFAULT, stdout=DEFAULT, argv=[
        'dummy.py',
        '--source_width', '0',
        '--traced_file', EXAMPLE_SOURCE_PATH,
        '--traced', 'example_source',
        '-m',
        'example_driver',
        '99'])
    def test_driver_module(self, stdin, stdout):
        source = """\
import sys
def foo(x):
    return sys.argv[1:]
"""
        expected_report = """\

x = 42
return ['99']
"""
        stdin.read.return_value = source

        main()

        report = stdout.write.call_args_list[0][0][0]
        self.assertReportEqual(expected_report, report)

    @patch.multiple('sys', stdin=DEFAULT, stdout=DEFAULT, argv=[
        'dummy.py',
        '--source_width', '0',
        '--traced_file', EXAMPLE_LIB_PATH,
        '--traced', 'example_package.lib_in_package',
        '-m',
        'example_driver'])
    def test_lib_in_package(self, stdin, stdout):
        source = """\
def add_message(s):
    package = __package__
    return s + ' Received'
"""
        expected_report = """\
s = 'from driver'
package = 'example_package'
return 'from driver Received'
"""
        stdin.read.return_value = source

        main()

        report = stdout.write.call_args_list[0][0][0]
        self.assertReportEqual(expected_report, report)

    @patch.multiple('sys', stdin=DEFAULT, stdout=DEFAULT, argv=[
        'dummy.py',
        '--source_width', '0',
        '--traced_file', EXAMPLE_SOURCE_PATH,
        '--traced', 'example_source',
        '-m',
        'example_package.driver_in_package'])
    def test_driver_in_package(self, stdin, stdout):
        source = """\
def foo(x):
    return 42
"""
        expected_report = """\
x = 'from driver in package'
return 42
"""
        stdin.read.return_value = source

        main()

        report = stdout.write.call_args_list[0][0][0]
        self.assertReportEqual(expected_report, report)

    # noinspection DuplicatedCode
    @patch.multiple('sys', stdin=DEFAULT, stdout=DEFAULT, argv=[
        'dummy.py',
        '--source_width', '0',
        '--traced_file', EXAMPLE_SOURCE_PATH,
        '--traced', 'example_source',
        EXAMPLE_DRIVER_PATH,
        'fail',
        'badly'])
    def test_driver_fails(self, stdin, stdout):
        source = """\
foo = 'Hello, World!'
"""
        expected_report = """\
foo = 'Hello, World!' | ---------------------------------------------------- |
                      | Traceback (most recent call last):                   |
                      |   File "path/example_driver.py", line 6, in <module> |
                      |     assert 'fail' not in sys.argv, sys.argv[1:]      |
                      | AssertionError: ['fail', 'badly']                    |
                      | ---------------------------------------------------- |
"""

        stdin.read.return_value = source

        with self.assertRaises(SystemExit):
            main()

        report = stdout.write.call_args_list[0][0][0]
        report = self.trim_exception(report)
        expected_report = self.trim_exception(expected_report)
        self.assertReportEqual(expected_report, report)

    # noinspection DuplicatedCode
    @patch.multiple('sys', stdin=DEFAULT, stdout=DEFAULT, argv=[
        'dummy.py',
        '--source_width', '0',
        '--traced_file', EXAMPLE_SOURCE_PATH,
        '--traced', 'example_source',
        EXAMPLE_PYCHARM_FAILURES_PATH])
    def test_driver_pycharm_failures(self, stdin, stdout):
        """ PyCharm's Pytest wrapper reports failures, but doesn't set exit code.

        Look for === FAILURES === report.
        """
        source = """\
foo = 'Hello, World!'
"""
        expected_report = """\
foo = 'Hello, World!' | ------------------------- |
                      | Pytest reported failures. |
                      | ------------------------- |
"""

        stdin.read.return_value = source

        with self.assertRaises(SystemExit):
            main()

        report = stdout.write.call_args_list[0][0][0]
        report = self.trim_exception(report)
        expected_report = self.trim_exception(expected_report)
        self.assertReportEqual(expected_report, report)

    # noinspection DuplicatedCode
    @patch.multiple('sys', stdin=DEFAULT, stdout=DEFAULT, argv=[
        'dummy.py',
        '--traced_file', 'foo.py',
        'bogus_driver.py'])
    def test_unknown_driver(self, stdin, stdout):
        source = """\
s = 'Yo!'
"""
        expected_report = """\

FileNotFoundError: [Errno 2] No such file or directory: 'bogus_driver.py' |
"""

        stdin.read.return_value = source

        with self.assertRaises(SystemExit):
            main()

        report = stdout.write.call_args_list[0][0][0]
        report = self.trim_exception(report)
        expected_report = self.trim_exception(expected_report)
        self.assertReportEqual(expected_report, report)

    @patch.multiple('sys', stdin=DEFAULT, stdout=DEFAULT, argv=[
        'dummy.py',
        '--source_width', '0',
        '--traced_file', 'different_source.py',
        '--traced', 'different_source',
        EXAMPLE_DRIVER_PATH])
    def test_bad_driver(self, stdin, stdout):
        source = """\
def foo(x):
    name = __name__
    return x + 1

BAR = 'baz'
"""
        expected_report = """\
----------------------------------------------------------------------------------- |
example_driver.py doesn't call the different_source module. Try a different driver. |
----------------------------------------------------------------------------------- |







"""
        stdin.read.return_value = source

        main()

        report = stdout.write.call_args_list[0][0][0]
        self.assertReportEqual(expected_report, report)

    @patch.multiple('sys', stdin=DEFAULT, stdout=DEFAULT, argv=[
        'dummy.py',
        '--source_width', '0',
        '--traced_file', EXAMPLE_SOURCE_PATH,
        '--traced', 'example_source',
        EXAMPLE_PATCHING_DRIVER_PATH])
    def test_driver_imports_first(self, stdin, stdout):
        source = """\
# This will raise a TypeError, unless we patch the sum() function before
# importing this module. example_patching_driver.py does the patch, so
# it has to be imported before this module.
start = sum([1, 2, "3"])

def foo(x):
    return x + start
"""
        expected_report = """\



start = 99

x = 10
return 109
"""
        stdin.read.return_value = source

        try:
            main()
        except SystemExit:
            pass

        report = stdout.write.call_args_list[0][0][0]
        self.assertReportEqual(expected_report, report)

    @patch.multiple('sys', stdin=DEFAULT, stdout=DEFAULT, argv=[
        'dummy.py',
        '--bad_driver', "Run config 'example' is bad, try something else.",
        '--source_width', '0',
        '--traced_file', 'different_source.py',
        '--traced', 'different_source',
        EXAMPLE_DRIVER_PATH])
    def test_bad_driver_message(self, stdin, stdout):
        source = """\
def foo(x):
    name = __name__
    return x + 1

BAR = 'baz'
"""
        expected_report = """\
------------------------------------------------ |
Run config 'example' is bad, try something else. |
------------------------------------------------ |







"""
        stdin.read.return_value = source

        main()

        report = stdout.write.call_args_list[0][0][0]
        self.assertReportEqual(expected_report, report)

    @patch.multiple('sys', stdin=DEFAULT, stdout=DEFAULT, argv=[
        'dummy.py',
        '--source_width', '0',
        '--traced_file', 'foo.py',
        '--traced', 'foo',
        '-m',
        'unittest',
        'foo'])
    def test_unittest_driver_passes(self, stdin, stdout):
        source = """\
from unittest import TestCase

def get_foo(x):
    return x + 5

class FooTest(TestCase):
    def test_get_foo(self):
        y = get_foo(10)
        self.assertEqual(15, y)
"""
        expected_report = """\


x = 10
return 15



y = 15
"""

        stdin.read.return_value = source

        main()

        report = stdout.write.call_args_list[0][0][0]
        report = self.trim_exception(report)
        expected_report = self.trim_exception(expected_report)
        self.assertReportEqual(expected_report, report)

    @patch.multiple('sys', stdin=DEFAULT, stdout=DEFAULT, argv=[
        'dummy.py',
        '--source_width', '0',
        '--traced_file', 'foo.py',
        '--traced', 'foo',
        '-m',
        'unittest',
        'foo'])
    def test_unittest_driver_fails(self, stdin, stdout):
        source = """\
from unittest import TestCase

def get_foo(x):
    return x + 500

class FooTest(TestCase):
    def test_get_foo(self):
        y = get_foo(10)
        self.fail(y)
"""
        expected_report = """\
---------------- |
SystemExit: True |
---------------- | | x = 10
                   | return 510



y = 510
AssertionError: 510
"""
        if sys.version_info < (3, 0):
            expected_report = expected_report.replace('(failures=1)',
                                                      'FAIL        ')

        stdin.read.return_value = source

        with self.assertRaises(SystemExit) as ctx:
            main()

        # noinspection PyUnresolvedReferences
        self.assertEqual(1, ctx.exception.code)
        report = stdout.write.call_args_list[0][0][0]
        report = self.trim_exception(report)
        expected_report = self.trim_exception(expected_report)
        self.assertReportEqual(expected_report, report)

    @patch.multiple('sys', stdin=DEFAULT, stdout=DEFAULT, argv=[
        'dummy.py',
        '--source_width', '0',
        '--traced_file', EXAMPLE_SOURCE_PATH,
        '--traced', 'example_source',
        '-m',
        'unittest',
        'example_silent_driver'])
    def test_silent_driver(self, stdin, stdout):
        """ Driver calls code, but doesn't generate messages. """
        source = """\
def bar(bucket):
    bucket.add('bar')
"""
        expected_report = """\

"""

        stdin.read.return_value = source

        main()

        report = stdout.write.call_args_list[0][0][0]
        report = self.trim_exception(report)
        expected_report = self.trim_exception(expected_report)
        self.assertReportEqual(expected_report, report)

    @patch.multiple('sys', stdin=DEFAULT, stdout=DEFAULT, argv=[
        'dummy.py',
        '--source_width', '0',
        '--traced_file', 'foo.py',
        '--traced', 'foo',
        '-m',
        'doctest',
        'foo.py'])
    def test_doctest_driver_fails(self, stdin, stdout):
        source = """\




def get_foo(x):
    ''' Example for doctest.
    
    >>> get_foo(42)
    942
    '''
    return x + 500
"""
        expected_report = """\
------------------------------------------------ |
SystemExit: 1                                    |
------------------------------------------------ |

x = 42





return 542
"""

        stdin.read.return_value = source
        stdout.encoding = None

        with self.assertRaises(SystemExit):
            main()

        report = stdout.write.call_args_list[0][0][0]
        expected_report = self.trim_exception(expected_report)
        report = self.trim_exception(report)
        self.assertReportEqual(expected_report, report)

    @staticmethod
    def trim_exception(report):
        report = re.sub(r"([ -])+\| *$", "", report, flags=re.MULTILINE)
        report = re.sub(r"line \d+", "line 9999", report)
        report = report.replace("IOError", "FileNotFoundError")
        report = report.replace('path/example_driver.py', EXAMPLE_DRIVER_PATH)
        report = report.replace('path/doctest.py',
                                str(doctest.__file__).strip('c'))
        return report

    @patch.multiple('sys', stdin=DEFAULT, stdout=DEFAULT, argv=[
        'dummy.py',
        '--source_width', '0',
        '--traced_file', EXAMPLE_SOURCE_PATH,
        '--traced', 'example_source',
        '-m',
        'example_package'])
    def test_driver_package(self, stdin, stdout):
        source = """\
def foo(x):
    return 42
"""
        expected_report = """\
x = 'from package __main__.py'
return 42
"""
        stdin.read.return_value = source

        main()

        report = stdout.write.call_args_list[0][0][0]
        self.assertReportEqual(expected_report, report)


def test_dunder_file(stdin, stdout, argv):
    argv.extend([
        'dummy.py',
        '--source_width', '0',
        '--traced_file', EXAMPLE_SOURCE_PATH,
        EXAMPLE_SOURCE_PATH])
    source = """\
import os

filename = os.path.basename(__file__)
"""
    expected_report = """\


filename = 'example_source.py'
"""
    stdin.read.return_value = source

    main()

    assert expected_report == stdout.getvalue()


def test_dunder_file_for_module(stdin, stdout, argv):
    argv.extend([
        'dummy.py',
        '--source_width', '0',
        '--traced', 'example_source',
        '--traced_file', EXAMPLE_SOURCE_PATH,
        '-m', 'example_source'])
    source = """\
import os

filename = os.path.basename(__file__)
"""
    expected_report = """\


filename = 'example_source.py'
"""
    stdin.read.return_value = source

    main()

    assert expected_report == stdout.getvalue()


def test_canvas_main(stdin, stdout, argv):
    argv.extend([
        'dummy.py',
        '--source_width', '0',
        '--traced_file', EXAMPLE_SOURCE_PATH,
        '--canvas'])
    source = """\
from turtle import *
forward(100)
"""
    expected_report = """\
start_canvas
create_line
    400
    300
    500
    300
    fill='black'
    pensize=1
end_canvas
.


"""
    stdin.read.return_value = source

    main()

    assert expected_report == stdout.getvalue()


def test_canvas_error(stdin, stdout, argv):
    argv.extend([
        'dummy.py',
        '--source_width', '0',
        '--traced_file', EXAMPLE_SOURCE_PATH,
        '--canvas'])
    source = """\
1/0
"""
    stdin.read.return_value = source

    with pytest.raises(SystemExit):
        main()

    report = stdout.getvalue()
    report_lines = report.splitlines()
    end = report_lines.index('end_canvas')
    assert report_lines[end-1] == "    text='ZeroDivisionError: division by zero'"


def test_canvas_syntax_error(stdin, stdout, argv):
    argv.extend([
        'dummy.py',
        '--source_width', '0',
        '--traced_file', EXAMPLE_SOURCE_PATH,
        '--canvas'])
    source = """\
x = 1
    y = 2
"""
    stdin.read.return_value = source

    with pytest.raises(SystemExit):
        main()

    report = stdout.getvalue()
    report_lines = report.splitlines()
    end = report_lines.index('end_canvas')
    assert report_lines[end-1] == "    text='IndentationError: unexpected indent'"


def test_exception_with_driver(stdin, stdout, argv):
    argv.extend([
        'dummy.py',
        '--source_width', '0',
        '--traced_file', EXAMPLE_SOURCE_PATH,
        '--traced', 'example_source',
        EXAMPLE_DRIVER_PATH])
    source = """\
import sys
def foo(x):
    sys.exit('Bad stuff.')
"""
    expected_report = """\
---------------------- |
SystemExit: Bad stuff. | x = 42
---------------------- | SystemExit: Bad stuff.
"""
    stdin.read.return_value = source

    with pytest.raises(SystemExit):
        main()

    assert expected_report == stdout.getvalue()


def test_syntax_error(stdin, stdout, argv):
    argv.extend([
        'dummy.py',
        '--source_width', '0',
        '--traced_file', EXAMPLE_SOURCE_PATH,
        EXAMPLE_SOURCE_PATH])
    source = """\
def missing_body():
"""
    expected_report = """\
SyntaxError: unexpected EOF while parsing
"""
    stdin.read.return_value = source

    with pytest.raises(SystemExit):
        main()

    assert expected_report == stdout.getvalue()


def test_driver_syntax_error(stdin, stdout, argv):
    argv.extend([
        'dummy.py',
        '--source_width', '0',
        '--traced_file', EXAMPLE_SOURCE_PATH,
        '--traced', 'example_source',
        EXAMPLE_DRIVER_SYNTAX_ERROR_PATH])
    source = """\
x = 'Hello, World!'
"""
    expected_report = """\
{} line 4: SyntaxError: invalid syntax
""".format(EXAMPLE_DRIVER_SYNTAX_ERROR_PATH)
    stdin.read.return_value = source

    with pytest.raises(SystemExit):
        main()

    assert expected_report == stdout.getvalue()


def test_trace_default_module(stdout, argv):
    argv.extend([
        'dummy.py',
        EXAMPLE_PRINTING_PATH])
    expected_report = """\
from __future__ import print_function   |
                                        |
                                        |
def custom_print(text, suffix):         | text = 'Hello, example' | suffix = '!'
    print(text + suffix)                | print('Hello, example!')
                                        |
                                        |
if __name__ == '__main__':              |
    custom_print('Hello, example', '!') |
"""

    main()

    assert expected_report == stdout.getvalue()


def test_dump_whole_file(stdout, argv):
    argv.extend([
        'dummy.py',
        '--traced', '__main__',
        EXAMPLE_PRINTING_PATH])
    expected_report = """\
from __future__ import print_function   |
                                        |
                                        |
def custom_print(text, suffix):         | text = 'Hello, example' | suffix = '!'
    print(text + suffix)                | print('Hello, example!')
                                        |
                                        |
if __name__ == '__main__':              |
    custom_print('Hello, example', '!') |
"""

    main()

    assert expected_report == stdout.getvalue()


def test_traced(stdout, argv):
    argv.extend([
        'dummy.py',
        '--traced', '__main__.custom_print',
        EXAMPLE_PRINTING_PATH])
    expected_report = """\
def custom_print(text, suffix): | text = 'Hello, example' | suffix = '!'
    print(text + suffix)        | print('Hello, example!')
"""

    main()

    assert expected_report == stdout.getvalue()


def test_no_driver(capsys):
    expected_error = ('space_tracer: error: one of the following arguments '
                      'are required: driver or traced_file')
    with pytest.raises(SystemExit):
        TraceRunner().trace_command(['space_tracer'])
    error = capsys.readouterr().err.splitlines()[-1]
    assert error == expected_error


def test_main_from_stdin(stdin):
    stdin.read.return_value = """\
print(40+2)
"""
    expected_report = """\
print(40+2) | print('42')"""

    report = TraceRunner().trace_command(['space_tracer', '-'])

    assert report == expected_report


def test_traced_file_without_traced():
    code = '''\
def foo(n):
    return n + 20
'''
    expected_report = '''\
n = 42
return 62'''

    with replace_input(code):
        report = TraceRunner().trace_command([
            'space_tracer',
            '--source_width', '0',
            '--traced_file', EXAMPLE_SOURCE_PATH,
            EXAMPLE_DRIVER_PATH])

    assert report == expected_report


def test_bad_driver_for_traced_file_without_traced():
    code = '''\
def foo(n):
    return n + 20
'''
    expected_report = '''\
--------------------------------------------------------------------------- |
example_driver.py doesn't call example_printing.py. Try a different driver. |
--------------------------------------------------------------------------- |







'''

    with replace_input(code):
        report = TraceRunner().trace_command([
            'space_tracer',
            '--traced_file', EXAMPLE_PRINTING_PATH,
            EXAMPLE_DRIVER_PATH])

    assert report == expected_report


def test_traced_driver_environment():
    code = '''\
try:
    sys.exit("Sys exists, but it wasn't imported!")
except NameError:
    pass
sys = None  # Would mess up driver script if they shared module name spaces.

def foo(x):
    return x + 20
'''
    expected_report = '''\

NameError: name 'sys' is not defined


sys = None

x = 42
return 62'''

    with replace_input(code):
        report = TraceRunner().trace_command([
            'space_tracer',
            '--source_width', '0',
            '--traced_file', EXAMPLE_SOURCE_PATH,
            '--traced', 'example_source',
            EXAMPLE_DRIVER_PATH])

    assert report == expected_report


def test_top_level_error():
    code = '''\
exit('Failed')
'''
    expected_report = '''\
SystemExit: Failed'''

    with replace_input(code):
        report = TraceRunner().trace_command([
            'space_tracer',
            '--source_width', '0',
            '--traced_file', EXAMPLE_DRIVER_PATH,
            '--traced=__main__',
            EXAMPLE_DRIVER_PATH])

    assert report == expected_report


def test_trace_child_package():
    expected_report = """\
def add_message(s):        | s = 'from driver'
    return s + ' received' | return 'from driver received'"""

    report = TraceRunner().trace_command([
        'space_tracer',
        '--traced=example_package.lib_in_package.add_message',
        EXAMPLE_DRIVER_PATH])

    assert report == expected_report



import re

import pytest
import sys

from space_tracer.main import TraceRunner
from test_report_builder import trim_report


@pytest.fixture(name='is_matplotlib_cleared')
def clear_matplotlib():
    for should_yield in (True, False):
        to_delete = [module_name
                     for module_name in sys.modules
                     if module_name.startswith('matplotlib')]
        for module_name in to_delete:
            del sys.modules[module_name]
        if should_yield:
            yield True


def replace_image(report):
    report = trim_report(report)
    report = re.sub(r"image='[a-zA-Z0-9+/=]*'", "image='...'", report)
    return report


def test_show(is_matplotlib_cleared):
    assert is_matplotlib_cleared
    code = """\
import matplotlib.pyplot as plt

data = [1, 2]
plt.plot(data)
plt.show()
"""
    expected_report = """\
create_image
    0
    0
    image='...'
"""
    tracer = TraceRunner()

    report = tracer.trace_turtle(code)

    assert expected_report == replace_image(report)


def test_numpy_random():
    code = """\
import numpy as np 

x = np.random.normal(size=3)
"""
    tracer = TraceRunner()

    original_report = tracer.trace_code(code)

    for _ in range(5):
        report = tracer.trace_code(code)
        assert original_report == report



import sys

from example_package.lib_in_package import add_message
from example_source import foo

assert 'fail' not in sys.argv, sys.argv[1:]

foo(42)
add_message('from driver')

sys.exit(0)  # Some drivers exit successfully like this.



from space_tracer.main import TraceRunner
from test_report_builder import trim_report


def test_repr():
    code = """\
class Dog(object):
    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return 'Dog(%r)' % self.name

dog = Dog('Spot')
animal = dog
"""
    expected_report = """\

name = 'Spot'
self.name = 'Spot'




dog = Dog('Spot')
animal = Dog('Spot') """
    tracer = TraceRunner()

    report = tracer.trace_code(code)

    assert trim_report(expected_report) == trim_report(report)


def test_repr_call():
    code = """\
class Dog(object):
    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return 'Dog(%r)' % self.name

dog1 = Dog('Spot')
dog2 = Dog('Fido')
s = repr(dog2)
"""
    expected_report = """\

name = 'Spot'      | name = 'Fido'
self.name = 'Spot' | self.name = 'Fido'


return "Dog('Fido')"

dog1 = Dog('Spot')
dog2 = Dog('Fido')
s = "Dog('Fido')"
"""
    tracer = TraceRunner()

    report = tracer.trace_code(code)

    assert trim_report(expected_report) == trim_report(report)


def test_repr_return():
    code = """\
class Foo:
    def __init__(self, x):
        self.x = x

    def __repr__(self):
        return 'Foo({!r})'.format(self.x)

def bar(x):
    return Foo(x)

y = bar(23)
"""
    expected_report = """\

x = 23
self.x = 23




x = 23
return Foo(23)

y = Foo(23)
"""
    tracer = TraceRunner()

    report = tracer.trace_code(code)

    assert trim_report(expected_report) == trim_report(report)


def test_repr_yield():
    code = """\
class Foo:
    def __init__(self, x):
        self.x = x

    def __repr__(self):
        return 'Foo({!r})'.format(self.x)

def bar(x):
    yield Foo(x)

y = list(bar(23))
"""
    expected_report = """\

x = 23
self.x = 23




x = 23
yield Foo(23)

y = [Foo(23)]
"""
    tracer = TraceRunner()

    report = tracer.trace_code(code)

    assert trim_report(expected_report) == trim_report(report)


def test_repr_yield_tuple():
    code = """\
class Foo:
    def __init__(self, x):
        self.x = x

    def __repr__(self):
        return 'Foo({!r})'.format(self.x)

def bar(x):
    yield x, Foo(x)

y = list(bar(23))
"""
    expected_report = """\

x = 23
self.x = 23




x = 23
yield (23, Foo(23))

y = [(23, Foo(23))]
"""
    tracer = TraceRunner()

    report = tracer.trace_code(code)

    assert trim_report(expected_report) == trim_report(report)


def test_repr_lambda():
    code = """\
class Foo:
    def __init__(self, x):
        self.x = x

    def __repr__(self):
        return 'Foo({!r})'.format(self.x)

y = list(map(lambda n: Foo(n), range(2)))
"""
    expected_report = """\

x = 0      | x = 1
self.x = 0 | self.x = 1




(0 => Foo(0)) | (1 => Foo(1)) | y = [Foo(0), Foo(1)] 
"""
    tracer = TraceRunner()

    report = tracer.trace_code(code)

    assert trim_report(expected_report) == trim_report(report)


def test_repr_mock():
    code = """\
from unittest.mock import Mock

m = Mock(name='foo')
"""
    expected_report = """\


m = Mock(name='foo')
"""
    tracer = TraceRunner()

    report = tracer.trace_code(code)

    assert trim_report(expected_report) == trim_report(report)


def test_repr_magic_mock():
    code = """\
from unittest.mock import MagicMock

m = MagicMock(name='foo')
"""
    expected_report = """\


m = MagicMock(name='foo')
"""
    tracer = TraceRunner()

    report = tracer.trace_code(code)

    assert trim_report(expected_report) == trim_report(report)


def test_repr_unnamed_mock():
    code = """\
from unittest.mock import Mock

m = Mock()
"""
    expected_report = """\


m = Mock()
"""
    tracer = TraceRunner()

    report = tracer.trace_code(code)

    assert trim_report(expected_report) == trim_report(report)


def test_repr_enum():
    code = """\
from enum import Enum

Animal = Enum('Animal', 'ANT BEE CAT DOG')
eric = Animal.BEE
"""
    expected_report = """\



eric = <Animal.BEE: 2>
"""
    tracer = TraceRunner()

    report = tracer.trace_code(code)

    assert trim_report(expected_report) == trim_report(report)


def test_error():
    code = """\
from __future__ import print_function

class Foo(object):
    def __repr__(self):
        raise RuntimeError('Bad representation.')

def create_foo():
    return Foo()

print('Start.')
foo = create_foo()
print('End.')
"""
    expected_report = """\







return <Foo object>

print('Start.')

print('End.')
"""
    tracer = TraceRunner()

    report = tracer.trace_code(code)

    assert trim_report(expected_report) == trim_report(report)



import os
import sys
from tempfile import TemporaryFile
from unittest import skipIf

from unittest.mock import patch

from space_tracer.main import TraceRunner, FileSwallower, replace_input
from space_tracer.mock_turtle import MockTurtle
from space_tracer.report_builder import ReportBuilder
from test_code_tracer_main import EXAMPLE_LIB_PATH
from test_report_builder import ReportTestCase


class CodeTracerTest(ReportTestCase):
    def setUp(self):
        super(CodeTracerTest, self).setUp()
        self.maxDiff = None

    def tearDown(self):
        MockTurtle.remove_monkey_patch()

    def test_empty(self):
        # EXEC
        report = TraceRunner().trace_code("")
        expected_report = ""

        # VERIFY
        self.assertReportEqual(expected_report, report)

    def test_chained_function(self):
        # SETUP
        code = """\
def foo(x):
    return x + 10

def bar(y):
    return foo(y) - 2

n = bar(3)
"""
        expected_report = """\
x = 3
return 13

y = 3
return 11

n = 11 """
        tracer = TraceRunner()

        # EXEC
        report = tracer.trace_code(code)

        # VERIFY
        self.assertReportEqual(expected_report, report)

    def test_call_on_literal(self):
        # SETUP
        code = """\
s = 'abc'.replace('a', 'A')
"""
        expected_report = """\
s = 'Abc' """
        tracer = TraceRunner()

        # EXEC
        report = tracer.trace_code(code)

        # VERIFY
        self.assertReportEqual(expected_report, report)

    def test_function_called_twice(self):
        # SETUP
        code = """\
def foo():
    x = 2
    return x + 10

n = foo()
r = foo()
"""
        expected_report = """\
          |
x = 2     | x = 2
return 12 | return 12

n = 12
r = 12 """
        tracer = TraceRunner()

        # EXEC
        report = tracer.trace_code(code)

        # VERIFY
        self.assertReportEqual(expected_report, report)

    def test_import(self):
        # SETUP
        code = """\
from decimal import Decimal

n = Decimal('10')
"""
        expected_report = """\


n = Decimal('10') """
        tracer = TraceRunner()

        # EXEC
        report = tracer.trace_code(code)

        # VERIFY
        self.assertReportEqual(expected_report, report)

    def test_compile_error(self):
        # SETUP
        code = """\
n = 1

if n > 0:
n -= 1
"""
        expected_report = """\



IndentationError: expected an indented block """
        tracer = TraceRunner()

        # EXEC
        report = tracer.trace_code(code)

        # VERIFY
        self.assertReportEqual(expected_report, report)

    def test_infinite_recursion_by_width(self):
        # SETUP
        code = """\
def foo(n):
    foo(n+1)

foo(0)
"""
        expected_report = """\
n = 0                                            | n = 1                                            |
RuntimeError: live coding message limit exceeded | RuntimeError: live coding message limit exceeded |

RuntimeError: live coding message limit exceeded
"""
        tracer = TraceRunner()
        tracer.max_width = 13

        # EXEC
        report = tracer.trace_code(code)

        # VERIFY
        self.assertReportEqual(expected_report, report)

    def test_infinite_loop_in_repr(self):
        # SETUP
        # __repr__() doesn't display messages, so message counter doesn't change.
        code = """\
class Foo(object):
    def __repr__(self):
        while True:
            x = 42

s = Foo()
s2 = 'x'
"""
        # The infinite loop fails silently inside __repr__().
        expected_report = """\






s2 = 'x'
"""
        tracer = TraceRunner()

        # EXEC
        report = tracer.trace_code(code)

        # VERIFY
        self.assertReportEqual(expected_report, report)

    def test_infinite_recursion_in_repr(self):
        # SETUP
        code = """\
class Foo:
    def get_bar(self):
        self.do_foo()
        return 'bar'

    def __repr__(self):
        return 'Foo' + self.get_bar()

    def do_foo(self):
        pass

f = Foo()
f.get_bar()
"""
        expected_report = """\



return 'bar'







f = Foobar
"""
        tracer = TraceRunner()

        # EXEC
        report = tracer.trace_code(code)

        # VERIFY
        self.assertReportEqual(expected_report, report)

    def test_set_attribute(self):
        # SETUP
        code = """\
class Dog(object):
    pass

dog = Dog()
dog.name = "Spot"
"""
        expected_report = """\




dog.name = 'Spot' """
        tracer = TraceRunner()

        # EXEC
        report = tracer.trace_code(code)

        # VERIFY
        self.assertReportEqual(expected_report, report)

    def test_set_attribute_item(self):
        # SETUP
        code = """\
class Shelf(object):
    pass

shelf = Shelf()
shelf.counts = {}
shelf.counts[2] = 3
"""
        expected_report = """\




shelf.counts = {}
shelf.counts[2] = 3 """
        tracer = TraceRunner()

        # EXEC
        report = tracer.trace_code(code)

        # VERIFY
        self.assertReportEqual(expected_report, report)

    def test_set_attribute_attribute(self):
        # SETUP
        code = """\
class foo(object):
    pass

f = foo()
f.child = foo()
f.child.name = 'bob'
"""
        expected_report = """\





f.child.name = 'bob' """
        tracer = TraceRunner()

        # EXEC
        report = tracer.trace_code(code)

        # VERIFY
        self.assertReportEqual(expected_report, report)

    def test_old_class(self):
        # SETUP
        code = """\
class Dog():
    pass

dog = Dog()
dog.name = "Spot"
"""
        expected_report = """\




dog.name = 'Spot' """
        tracer = TraceRunner()

        # EXEC
        report = tracer.trace_code(code)

        # VERIFY
        self.assertReportEqual(expected_report, report)

    def test_module_name(self):
        # SETUP
        code = """\
def foo(x):
    return x + 3

if __name__ == '__main__':
    y = foo(10)
"""
        expected_report = """\
x = 10
return 13


y = 13 """
        tracer = TraceRunner()

        # EXEC
        with replace_input(code):
            report = tracer.trace_command([
                'space_tracer',
                '--source_width', '0',
                '--traced_file', 'example.py',
                'example.py'])

        # VERIFY
        self.assertReportEqual(expected_report, report)

    def test_module_name_live(self):
        # SETUP
        code = """\
def foo(x):
    return x + 3

if __name__ == '__live_coding__':
    y = foo(10)
"""
        expected_report = """\
x = 10
return 13


y = 13 """
        tracer = TraceRunner()

        # EXEC
        report = tracer.trace_code(code)

        # VERIFY
        self.assertReportEqual(expected_report, report)

    def test_lambda(self):
        # SETUP
        code = """\
f = lambda n: n + 1
x = f(10)
"""
        expected_report = """\
(10 => 11)
x = 11 """
        tracer = TraceRunner()

        # EXEC
        report = tracer.trace_code(code)
        # VERIFY
        self.assertReportEqual(expected_report, report)

    def test_turtle(self):
        # SETUP
        code = """\
from turtle import *

forward(100)
"""
        expected_report = """\
create_line
    0
    0
    100
    0
    fill='black'
    pensize=1
"""
        tracer = TraceRunner()

        # EXEC
        report = tracer.trace_turtle(code)

        # VERIFY
        self.assertReportEqual(expected_report, report)

    def test_mainloop(self):
        # SETUP
        code = """\
from turtle import *

mainloop()
done()  # alias for mainloop()
"""
        expected_report = """\



"""
        tracer = TraceRunner()

        # EXEC
        report = tracer.trace_code(code)

        # VERIFY
        self.assertReportEqual(expected_report, report)

    def test_yield(self):
        # SETUP
        code = """\
def foo(x):
    a = x
    yield a
    a += 2
    yield a

n = foo(10)
s = sum(n)
"""
        expected_report = """\
x = 10
a = 10
yield 10
a = 12
yield 12


s = 22 """
        tracer = TraceRunner()

        # EXEC
        report = tracer.trace_code(code)

        # VERIFY
        self.assertReportEqual(expected_report, report)

    def test_method_of_anonymous_object(self):
        # SETUP
        code = """\
def make_string():
    return 'abc'

s = make_string().upper()
"""
        expected_report = """\

return 'abc'

s = 'ABC'
"""
        # EXEC
        report = TraceRunner().trace_code(code)

        # VERIFY
        self.assertReportEqual(expected_report, report)

    def test_recursion(self):
        # SETUP
        code = """\
def f(n):
    r = 1
    for i in range(n):
        r += f(i)
    return r

r = f(2)
"""
        expected_report = """\
n = 2         | n = 0    | n = 1    | n = 0
r = 1         | r = 1    | r = 1    | r = 1
i = 0 | i = 1 |          | i = 0    |
r = 2 | r = 4 |          | r = 2    |
return 4      | return 1 | return 2 | return 1

r = 4 """
        tracer = TraceRunner()

        # EXEC
        report = tracer.trace_code(code)

        # VERIFY
        self.assertReportEqual(expected_report, report)

    def test_recursion_exception(self):
        # SETUP
        code = """\
def f(n):
    m = n - 1
    if m == 0:
        raise RuntimeError('Invalid n.')
    return f(m)

r = f(2)
"""
        expected_report = """\
n = 2                    | n = 1
m = 1                    | m = 0
                         |
                         | RuntimeError: Invalid n.
RuntimeError: Invalid n. |

RuntimeError: Invalid n. """
        tracer = TraceRunner()

        # EXEC
        report = tracer.trace_code(code)

        # VERIFY
        self.assertReportEqual(expected_report, report)

    def test_return_tuple(self):
        # SETUP
        code = """\
def f():
    return (1, 2)

x = f()
"""
        expected_report = """\

return (1, 2)

x = (1, 2) """
        # EXEC
        report = TraceRunner().trace_code(code)

        # VERIFY
        self.assertReportEqual(expected_report, report)

    def test_multiline_exception(self):
        # SETUP
        code = """\
x = (
     '')
y = 2
raise RuntimeError('Bad stuff happened.')
"""
        expected_report = """\
x = ''

y = 2
RuntimeError: Bad stuff happened. """
        # EXEC
        report = TraceRunner().trace_code(code)

        # VERIFY
        self.assertReportEqual(expected_report, report)

    def test_yield_none(self):
        # SETUP
        code = """\
def foo():
    yield

for x in foo():
    pass
"""
        expected_report = """\

yield None

x = None """

        # EXEC
        report = TraceRunner().trace_code(code)

        # VERIFY
        self.assertReportEqual(expected_report, report)

    def test_yield_from(self):
        # SETUP
        code = """\
def foo(n):
    yield 10 + n
    yield 20 + n

def bar():
    for i in range(2):
        yield from foo(i)

for x in bar():
    pass
"""
        expected_report = """\
n = 0    | n = 1
yield 10 | yield 11
yield 20 | yield 21


i = 0               | i = 1
yield 10 | yield 20 | yield 11 | yield 21

x = 10 | x = 20 | x = 11 | x = 21
       |        |        |
"""

        # EXEC
        report = TraceRunner().trace_code(code)

        # VERIFY
        self.assertReportEqual(expected_report, report)

    def test_yield_with_return(self):
        # SETUP
        code = """\
def foo():
    yield 1
    return

for x in foo():
    pass
"""
        expected_report = """\

yield 1


x = 1 """

        # EXEC
        report = TraceRunner().trace_code(code)

        # VERIFY
        self.assertReportEqual(expected_report, report)

    def test_future(self):
        code = """\
from __future__ import print_function

print('x')
"""
        expected_report = """\


print('x') """

        report = TraceRunner().trace_code(code)

        self.assertReportEqual(expected_report, report)

    def test_future_after_docstring_and_comment(self):
        code = """\
''' My module '''

# some comment

from __future__ import print_function

print('x')
"""
        expected_report = """\






print('x') """

        report = TraceRunner().trace_code(code)

        self.assertReportEqual(expected_report, report)

    def test_display_star_args(self):
        code = """\
def foo(*args):
    pass

foo('Bob', 1)
"""
        expected_report = """\
args = ('Bob', 1)


"""

        report = TraceRunner().trace_code(code)

        self.assertReportEqual(expected_report, report)

    def test_display_kwargs(self):
        code = """\
def foo(**kwargs):
    pass

foo(name='Bob')
"""
        expected_report = """\
kwargs = {'name': 'Bob'}


"""

        report = TraceRunner().trace_code(code)

        self.assertReportEqual(expected_report, report)

    def test_delete_item(self):
        code = """\
l = [0, 10, 20]
del l[1]
"""
        expected_report = """\
l = [0, 10, 20]
l = [0, 20] """

        report = TraceRunner().trace_code(code)

        self.assertReportEqual(expected_report, report)

    def test_delete_attribute(self):
        code = """\
class Foo(object):
    def __init__(self, name=None):
        self.name = name

    def __repr__(self):
        name = getattr(self, 'name', None)
        if name is None:
            return 'Foo()'
        return 'Foo({!r})'.format(name)

f = Foo('Bob')
del f.name
g = f
"""
        expected_report = """\

name = 'Bob'
self.name = 'Bob'







f = Foo('Bob')
f = Foo()
g = Foo() """

        report = TraceRunner().trace_code(code)

        self.assertReportEqual(expected_report, report)

    def test_delete_item_of_attribute(self):
        code = """\
class Foo(object):
    pass
f = Foo()
f.l = [0, 10, 20]
del f.l[1]
"""
        expected_report = """\



f.l = [0, 10, 20]
f.l = [0, 20] """

        report = TraceRunner().trace_code(code)

        self.assertReportEqual(expected_report, report)

    @skipIf(sys.version_info < (3, 6),
            'f-strings not supported before Python 3.6.')
    def test_f_string(self):
        code = """\
x = 42
s = f'The answer is {x}.'
"""
        expected_report = """\
x = 42
s = 'The answer is 42.' """

        report = TraceRunner().trace_code(code)

        self.assertReportEqual(expected_report, report)

    @skipIf(sys.version_info < (3, 6),
            'f-strings not supported before Python 3.6.')
    def test_f_string_in_function(self):
        code = """\
y = 42

def foo(x):
    return f'The answer is {x}.'

z = foo(y)
"""
        expected_report = """\
y = 42

x = 42
return 'The answer is 42.'

z = 'The answer is 42.' """

        report = TraceRunner().trace_code(code)

        self.assertReportEqual(expected_report, report)

    @skipIf(sys.version_info < (3, 6),
            'f-strings not supported before Python 3.6.')
    def test_simple_f_string_in_function(self):
        code = """\
y = 42

def foo(x):
    return f'{x}'

z = foo(y)
"""
        expected_report = """\
y = 42

x = 42
return '42'

z = '42' """

        report = TraceRunner().trace_code(code)

        self.assertReportEqual(expected_report, report)

    def test_random(self):
        code = """\
from random import randint

i = randint(1, 100)
j = randint(1, 100)
"""
        if sys.version_info < (3, 0):
            expected_report = """\


i = 85
j = 76
"""
        else:
            expected_report = """\


i = 50
j = 98
"""

        report = TraceRunner().trace_code(code)

        self.assertReportEqual(expected_report, report)

    def test_builtins(self):
        code = """\
x = __builtins__.sum([1, 2])
"""
        expected_report = """\
x = 3
"""

        report = TraceRunner().trace_code(code)

        self.assertReportEqual(expected_report, report)

    def test_module_in_package(self):
        code = """\
from unittest import TestCase


class FooTest(TestCase):
    def test_foo(self):
        s = __package__
        self.assertEqual('example_package', s)
"""
        expected_report = """\





s = 'example_package'
"""

        with replace_input(code):
            report = TraceRunner().trace_command([
                'space_tracer',
                '--source_width', '0',
                '--traced', 'example_package.lib_in_package',
                '--traced_file', EXAMPLE_LIB_PATH,
                '-m',
                'unittest',
                'example_package.lib_in_package'])

        self.assertReportEqual(expected_report, report)

    def test_module_is_own_driver(self):
        code = """\
s = __package__
s = __name__
"""
        expected_report = """\
s = 'example_package'
s = '__live_coding__'
"""

        foo_path = os.path.join(os.path.dirname(__name__),
                                'example_package',
                                'foo.py')
        with replace_input(code):
            report = TraceRunner().trace_command([
                'space_tracer',
                '--source_width', '0',
                '--traced_file', foo_path,
                '--traced', '__live_coding__',
                '--live',
                '-m',
                'example_package.foo'])

        self.assertReportEqual(expected_report, report)


class FileSwallowerTest(ReportTestCase):
    def test_temp_file(self):
        expected_contents = 'before\nline 1\nline 2\nafter\n'

        with TemporaryFile('w+') as real_file:
            real_file.write('before\n')

            swallower = FileSwallower(real_file)
            swallower.write('line 1\n')
            swallower.write('line 2\n')

            real_file.write('after\n')
            real_file.seek(0)
            real_contents = real_file.read()

        self.assertEqual(expected_contents, real_contents)

    def test_buffer(self):
        expected_contents = 'before\nline 1\nline 2\nafter\n'

        with TemporaryFile('w+') as real_file:
            real_file.write('before\n')
            real_file.flush()

            swallower = FileSwallower(real_file)
            swallower_buffer = getattr(swallower, 'buffer', swallower)
            swallower_buffer.write(b'line 1\n')
            swallower_buffer.write(b'line 2\n')

            real_file.write('after\n')
            real_file.seek(0)
            real_contents = real_file.read()

        self.assertEqual(expected_contents, real_contents)

    def test_report(self):
        source = """\
from __future__ import print_function
import example_printing
__live_coding_context__ = globals()['report_builder']  # variable name is important!
example_printing.custom_print('42', 'xyz')
"""
        expected_report = """\



print('42xyz')
"""
        report_builder = ReportBuilder()
        environment = dict(report_builder=report_builder)
        swallower = FileSwallower(sys.stdout)

        with patch('sys.stdout', swallower):
            exec(source, environment, environment)
        report = report_builder.report()

        self.assertReportEqual(expected_report, report)

    def test_long_values(self):
        code = """\
y = [x for x in range(100)]
"""
        expected_report = """\
y = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, [320 chars]91, 92, 93, 94, 95, 96, 97, 98, 99]
"""
        report = TraceRunner().trace_code(code)
        self.assertReportEqual(expected_report, report)

    def test_long_values_in_function(self):
        self.maxDiff = None
        code = """\
def main(): 
    for i in range(2): 
        x = 100 * (i+1) 
        y = ['x'] * x 
 
main() 
"""
        expected_report = """\

i = 0                                                                                 | i = 1 
x = 100                                                                               | x = 200 
y = ['x', 'x', 'x', 'x', 'x', 'x', 'x',[430 chars] 'x', 'x', 'x', 'x', 'x', 'x', 'x'] | y = ['x', 'x', 'x', \
'x', 'x', 'x', 'x',[930 chars] 'x', 'x', 'x', 'x', 'x', 'x', 'x']

"""
        report = TraceRunner().trace_code(code)
        self.assertReportEqual(expected_report, report)



import os
import sys
from subprocess import check_output, CalledProcessError

if sys.version_info >= (3, 0):
    print('This is modern Python. Nothing to validate.')
    exit()

source_path = os.path.abspath(os.path.join(__file__, '../example_printing.py'))
lib_path = os.path.abspath(os.path.join(__file__, '../../../../plugin/PySrc'))
args = [sys.executable, '-mspace_tracer', source_path]

try:
    check_output(args, cwd=lib_path).decode('utf8')
    exit('Space Tracer should fail under Python 2.')
except CalledProcessError as ex:
    assert ex.output == 'Python 2 is no longer supported.\n', ex.output

print('OK')



import sys

from space_tracer.main import TraceRunner
from test_report_builder import trim_report


def test_assign_tuple():
    code = """\
from __future__ import print_function

a = (1,)
b, = (1,)
(c) = (1,)
(d,) = (1,)

print(a)
print(b)
print(c)
print(d)

a, b = 1, 2
(c, d) = (3, 4)
"""
    expected_report = """\


a = (1,)
(b,) = (1,)
c = (1,)
(d,) = (1,)

print('(1,)')
print('1')
print('(1,)')
print('1')

(a, b) = (1, 2)
(c, d) = (3, 4)
"""
    tracer = TraceRunner()

    report = tracer.trace_code(code)

    assert expected_report == trim_report(report)


def test_assign_tuple_tuple():
    code = """\
a, (b, c) = (1, (2, 3))
"""
    expected_report = """\
(a, (b, c)) = (1, (2, 3))
"""
    tracer = TraceRunner()

    report = tracer.trace_code(code)

    assert expected_report == trim_report(report)


def test_assign_tuple_list():
    code = """\
a, [b, c] = (1, (2, 3))
"""
    expected_report = """\
(a, (b, c)) = (1, (2, 3))
"""
    tracer = TraceRunner()

    report = tracer.trace_code(code)

    assert expected_report == trim_report(report)


def test_assign_generator_unpacked():
    code = """\
a, b = (3*i for i in range(2))
"""
    expected_report = """\
(a, b) = (0, 3)
"""
    tracer = TraceRunner()

    report = tracer.trace_code(code)

    assert expected_report == trim_report(report)


def test_assign_assignment():
    code = """\
a = b = 2
"""
    expected_report = """\
a = b = 2
"""
    tracer = TraceRunner()

    report = tracer.trace_code(code)

    assert expected_report == trim_report(report)


def test_assign_generator_assignment():
    """ Can't convert to tuple for more than one assignment. """
    code = """\
a, b = c = (3*i for i in range(2))
d = a, b
e = list(c)
"""
    expected_report = """\

d = (0, 3)
e = []
"""
    tracer = TraceRunner()

    report = tracer.trace_code(code)

    assert expected_report == trim_report(report)


def test_assign_to_anonymous_attribute():
    code = """\
class Foo(object):
    pass

Foo().x = 2
"""
    expected_report = """\



"""
    report = TraceRunner().trace_code(code)

    assert expected_report == trim_report(report)


def test_assign_to_expression():
    code = """\
a = [1, 2, 3]

(a or None)[1] = 20
"""
    expected_report = """\
a = [1, 2, 3]

"""
    report = TraceRunner().trace_code(code)

    assert expected_report == trim_report(report)


def test_augmented_assign_to_expression():
    code = """\
a = [1, 2, 3]

(a or None)[1] += 20
"""
    expected_report = """\
a = [1, 2, 3]

"""
    report = TraceRunner().trace_code(code)

    assert expected_report == trim_report(report)


def test_assign_starred():
    code = """\
a, *b = (1, 2, 3)
print(b)
"""
    expected_report = """\
(a, *b) = (1, 2, 3)
print('[2, 3]')
"""
    tracer = TraceRunner()

    report = tracer.trace_code(code)

    assert expected_report == trim_report(report)


def test_docstring():
    code = """\
''
"""
    expected_report = """\

"""
    tracer = TraceRunner()

    report = tracer.trace_code(code)

    assert expected_report == trim_report(report)


def test_runtime_error():
    code = """\
x = 2
raise RuntimeError('Bad stuff happened.')
"""
    expected_report = """\
x = 2
RuntimeError: Bad stuff happened.
"""
    tracer = TraceRunner()

    report = tracer.trace_code(code)

    assert expected_report == trim_report(report)


def test_runtime_error_after_conditional():
    code = """\
if False:
    x = 2
else:
    x = 3
raise RuntimeError('Bad stuff happened.')
"""
    expected_report = """\



x = 3
RuntimeError: Bad stuff happened.
"""
    tracer = TraceRunner()

    report = tracer.trace_code(code)

    assert expected_report == trim_report(report)


def test_runtime_error_caught():
    code = """\
try:
    raise RuntimeError('Bad stuff happened.')
except Exception as e:
    f = e
"""
    expected_report = """\

RuntimeError: Bad stuff happened.
e = RuntimeError('Bad stuff happened.',)
f = RuntimeError('Bad stuff happened.',)
"""
    if sys.version_info >= (3, 7, 0):
        expected_report = expected_report.replace(',)', ')')

    tracer = TraceRunner()

    report = tracer.trace_code(code)

    assert expected_report == trim_report(report)


def test_runtime_error_caught_unnamed():
    code = """\
try:
    raise RuntimeError('Bad stuff happened.')
except:
    f = 'Worse stuff'
"""
    expected_report = """\

RuntimeError: Bad stuff happened.

f = 'Worse stuff'
"""

    tracer = TraceRunner()

    report = tracer.trace_code(code)

    assert expected_report == trim_report(report)


def test_multiline_error():
    code = """\
quality = 0
for c in ['1',
          'x']:
    quality += int(c)
"""
    expected_report = """\
quality = 0
c = '1'     | c = 'x'
            |
quality = 1 | ValueError: invalid literal for int() with base 10: 'x'
"""
    tracer = TraceRunner()

    report = tracer.trace_code(code)

    assert expected_report == trim_report(report)


def test_unwinding_exceptions():
    code = """\
def foo(n):
    raise RuntimeError('Bad stuff happened.')

x = foo(5)
"""
    expected_report = """\
n = 5
RuntimeError: Bad stuff happened.

RuntimeError: Bad stuff happened.
"""

    tracer = TraceRunner()

    report = tracer.trace_code(code)

    assert expected_report == trim_report(report)


def test_reraise():
    code = """\
x = 2
try:
    raise RuntimeError('Bad stuff happened.')
except:
    raise
"""
    expected_report = """\
x = 2

RuntimeError: Bad stuff happened.

RuntimeError: Bad stuff happened.
"""
    tracer = TraceRunner()

    report = tracer.trace_code(code)

    assert expected_report == trim_report(report)


def test_nested_reraise():
    code = """\
x = 2
try:
    try:
        raise RuntimeError('Bad stuff happened.')
    except:
        raise
except:
    raise
"""
    expected_report = """\
x = 2


RuntimeError: Bad stuff happened.

RuntimeError: Bad stuff happened.

RuntimeError: Bad stuff happened.
"""
    tracer = TraceRunner()

    report = tracer.trace_code(code)

    assert expected_report == trim_report(report)



import os

from space_tracer.main import TraceRunner, replace_input
from test_code_tracer_main import EXAMPLE_DRIVER_PATH, EXAMPLE_PRINTING_PATH, EXAMPLE_SOURCE_PATH
from test_report_builder import trim_report


def test_one_function():
    code = """\
from space_tracer import traced


@traced
def foo(n):
    s = 'x'
    for i in range(n):
        s += 'y'
    return s


def bar(num):
    s = 'a'
    for i in range(num):
        s += 'b'
    return s

print(foo(3))
print(bar(3))
"""
    expected_report = """\
    @traced                |
    def foo(n):            | n = 3
        s = 'x'            | s = 'x'
        for i in range(n): | i = 0    | i = 1     | i = 2
            s += 'y'       | s = 'xy' | s = 'xyy' | s = 'xyyy'
        return s           | return 'xyyy' """

    with replace_input(code):
        report = TraceRunner().trace_command([
            'space_tracer',
            '--source_indent', '4',
            '--traced_file', 'example.py',
            'example.py'])

    assert trim_report(expected_report) == trim_report(report)


def test_two_functions():
    code = """\
from space_tracer import traced


@traced
def foo(n):
    s = 'x'
    for i in range(n):
        s += 'y'
    return s


@traced
def bar(num):
    s = 'a'
    for i in range(num):
        s += 'b'
    return s


print(foo(3))
print(bar(3))
"""
    expected_report = """\
    @traced                  |
    def foo(n):              | n = 3
        s = 'x'              | s = 'x'
        for i in range(n):   | i = 0    | i = 1     | i = 2
            s += 'y'         | s = 'xy' | s = 'xyy' | s = 'xyyy'
        return s             | return 'xyyy'
    @traced                  |
    def bar(num):            | num = 3
        s = 'a'              | s = 'a'
        for i in range(num): | i = 0    | i = 1     | i = 2
            s += 'b'         | s = 'ab' | s = 'abb' | s = 'abbb'
        return s             | return 'abbb'
"""

    with replace_input(code):
        report = TraceRunner().trace_command([
            'space_tracer',
            '--source_indent', '4',
            '--traced_file', 'example.py',
            'example.py'])

    assert trim_report(expected_report) == trim_report(report)


def test_command_line():
    """ Specify a traced method with command-line option --traced. """
    code = """\
def foo(n):
    s = 'x'
    for i in range(n):
        s += 'y'
    return s


def bar(num):
    s = 'a'
    for i in range(num):
        s += 'b'
    return s


print(foo(3))
print(bar(3))
"""
    expected_report = """\
def bar(num):            | num = 3
    s = 'a'              | s = 'a'
    for i in range(num): | i = 0    | i = 1     | i = 2
        s += 'b'         | s = 'ab' | s = 'abb' | s = 'abbb'
    return s             | return 'abbb'"""

    with replace_input(code):
        report = TraceRunner().trace_command([
            'space_tracer',
            '--traced', '__main__.bar',
            '--traced_file', 'example.py',
            'example.py'])

    assert expected_report == report


def test_without_traced_file():
    expected_report = """\
def foo(x):                       | x = 42
    return x + 1                  | return 43
                                  |
                                  |
def bar(bucket):                  |
    bucket.add('bar')             |
                                  |
                                  |
if __name__ == '__live_coding__': |
    y = foo(3)                    |"""

    report = TraceRunner().trace_command([
        'space_tracer',
        '--traced', 'example_source',
        EXAMPLE_DRIVER_PATH])

    assert expected_report == report


def test_default_traced():
    expected_report = """\
from __future__ import print_function   |
                                        |
                                        |
def custom_print(text, suffix):         | text = 'Hello, example' | suffix = '!'
    print(text + suffix)                | print('Hello, example!')
                                        |
                                        |
if __name__ == '__main__':              |
    custom_print('Hello, example', '!') |"""

    report = TraceRunner().trace_command([
        'space_tracer',
        EXAMPLE_PRINTING_PATH])

    assert expected_report == report


def test_traced_main_without_traced_file():
    expected_report = """\
def custom_print(text, suffix): | text = 'Hello, example' | suffix = '!'
    print(text + suffix)        | print('Hello, example!')"""

    report = TraceRunner().trace_command([
        'space_tracer',
        '--traced=__main__.custom_print',
        EXAMPLE_PRINTING_PATH])

    assert expected_report == report


def test_traced_function():
    code = """\
def foo(n):
    return n + 1

def bar(n):
    return n - 1

foo(10)
bar(20)
"""
    expected_report = """\
def bar(n):      | n = 20
    return n - 1 | return 19"""

    with replace_input(code):
        report = TraceRunner().trace_command([
            'space_tracer',
            '--traced=bar',
            '--traced_file', 'example.py',
            'example.py'])

    assert expected_report == report


def test_other_decorator():
    """ Other decorators shouldn't affect tracing. """
    code = """\
from __future__ import print_function
class Foo(object):
    def foo(self, x):
        return x + 1
    
    @staticmethod
    def bar(x):
        return x + 2

f = Foo()
print(f.foo(10))
print(f.bar(20))
"""
    expected_report = """\
from __future__ import print_function |
class Foo(object):                    |
    def foo(self, x):                 | x = 10
        return x + 1                  | return 11
                                      |
    @staticmethod                     |
    def bar(x):                       | x = 20
        return x + 2                  | return 22
                                      |
f = Foo()                             |
print(f.foo(10))                      | print('11')
print(f.bar(20))                      | print('22')"""

    with replace_input(code):
        report = TraceRunner().trace_command([
            'space_tracer',
            '--traced_file', 'example.py',
            'example.py'])

    assert expected_report == report


def test_attribute_decorator():
    """ The decorator is a module attribute. """
    code = """\
from __future__ import print_function
class Foo(object):
    def foo(self, x):
        return x + 1
    
    @__builtins__.staticmethod
    def bar(x):
        return x + 2

f = Foo()
print(f.foo(10))
print(f.bar(20))
"""
    expected_report = """\
from __future__ import print_function |
class Foo(object):                    |
    def foo(self, x):                 | x = 10
        return x + 1                  | return 11
                                      |
    @__builtins__.staticmethod        |
    def bar(x):                       | x = 20
        return x + 2                  | return 22
                                      |
f = Foo()                             |
print(f.foo(10))                      | print('11')
print(f.bar(20))                      | print('22')"""

    with replace_input(code):
        report = TraceRunner().trace_command([
            'space_tracer',
            '--traced_file', 'example.py',
            'example.py'])

    assert expected_report == report


def test_traced_renamed():
    code = """\
from space_tracer import traced as space_traced


@space_traced
def foo(n):
    s = 'x'
    for i in range(n):
        s += 'y'
    return s

print(foo(3))
"""
    expected_report = """\
    @space_traced          |
    def foo(n):            | n = 3
        s = 'x'            | s = 'x'
        for i in range(n): | i = 0    | i = 1     | i = 2
            s += 'y'       | s = 'xy' | s = 'xyy' | s = 'xyyy'
        return s           | return 'xyyy' """

    with replace_input(code):
        report = TraceRunner().trace_command([
            'space_tracer',
            '--source_indent', '4',
            '--traced_file', 'example.py',
            'example.py'])

    assert trim_report(expected_report) == trim_report(report)


def test_traced_with_block():
    code = """\
from space_tracer import traced

def foo(n):
    s = 'x'
    with traced():
        t = 23
        for i in range(n):
            s += 'y'
    return s

print(foo(3))
"""
    expected_report = """\
    with traced():         |
        t = 23             | t = 23
        for i in range(n): | i = 0    | i = 1     | i = 2
            s += 'y'       | s = 'xy' | s = 'xyy' | s = 'xyyy'"""

    with replace_input(code):
        report = TraceRunner().trace_command([
            'space_tracer',
            '--traced_file', 'example.py',
            'example.py'])

    assert report == expected_report


def test_other_decorator_renamed():
    """ The decorator is a module attribute. """
    code = """\
traced = staticmethod

class Foo(object):
    def foo(self, x):
        return x + 1
    
    @traced
    def bar(x):
        return x + 2

f = Foo()
print(f.foo(10))
print(f.bar(20))
"""
    expected_report = """\
traced = staticmethod |
                      |
class Foo(object):    |
    def foo(self, x): | x = 10
        return x + 1  | return 11
                      |
    @traced           |
    def bar(x):       | x = 20
        return x + 2  | return 22
                      |
f = Foo()             |
print(f.foo(10))      | print('11')
print(f.bar(20))      | print('22')"""

    with replace_input(code):
        report = TraceRunner().trace_command([
            'space_tracer',
            '--traced_file', 'example.py',
            'example.py'])

    assert expected_report == report


def test_one_function_live_mode():
    """ Need to keep original vertical position to line up with editor. """
    code = """\
from space_tracer import traced


@traced
def foo(n):
    s = 'x'
    for i in range(n):
        s += 'y'
    return s


def bar(num):
    s = 'a'
    for i in range(num):
        s += 'b'
    return s


print(foo(3))
print(bar(3))
"""
    expected_report = """\




n = 3
s = 'x'
i = 0    | i = 1     | i = 2
s = 'xy' | s = 'xyy' | s = 'xyyy'
return 'xyyy'










"""

    with replace_input(code):
        report = TraceRunner().trace_command([
            'space_tracer',
            '--source_width', '0',
            '--live',
            '--traced_file', 'example.py',
            'example.py'])

    assert trim_report(expected_report) == trim_report(report)


def test_relative_traced_file():
    code = """\
def foo(x):
    return x + 100
"""
    expected_report = """\
def foo(x):        | x = 42
    return x + 100 | return 142"""
    relative_path = os.path.relpath(EXAMPLE_SOURCE_PATH, os.getcwd())

    with replace_input(code):
        report = TraceRunner().trace_command([
            'space_tracer',
            '--traced_file', relative_path,
            EXAMPLE_DRIVER_PATH])

    assert report == expected_report


def test_unknown_traced_file():
    code = 'Yo!'
    expected_report = """\
--------------------------------------------------------------------- |
example_driver.py doesn't call bogus_file.py. Try a different driver. |
--------------------------------------------------------------------- |







"""

    with replace_input(code):
        report = TraceRunner().trace_command([
            'space_tracer',
            '--traced_file', 'bogus_path/bogus_file.py',
            EXAMPLE_DRIVER_PATH])

    assert report == expected_report


def test_hide_name():
    code = """\
from space_tracer import traced


@traced(hide='i')
def foo(n):
    s = 'x'
    for i in range(n):
        s += 'y'
    return s

print(foo(3))
"""
    expected_report = """\
    @traced(hide='i')      |
    def foo(n):            | n = 3
        s = 'x'            | s = 'x'
        for i in range(n): |          |           |
            s += 'y'       | s = 'xy' | s = 'xyy' | s = 'xyyy'
        return s           | return 'xyyy'"""

    with replace_input(code):
        report = TraceRunner().trace_command([
            'space_tracer',
            '--source_indent', '4',
            '--traced_file', 'example.py',
            'example.py'])

    assert report == expected_report


def test_hide_names_list():
    code = """\
from space_tracer import traced


@traced(hide=['i', 'n'])
def foo(n):
    s = 'x'
    for i in range(n):
        s += 'y'
    return s

print(foo(3))
"""
    expected_report = """\
    @traced(hide=['i', 'n']) |
    def foo(n):              |
        s = 'x'              | s = 'x'
        for i in range(n):   |          |           |
            s += 'y'         | s = 'xy' | s = 'xyy' | s = 'xyyy'
        return s             | return 'xyyy'"""

    with replace_input(code):
        report = TraceRunner().trace_command([
            'space_tracer',
            '--source_indent', '4',
            '--traced_file', 'example.py',
            'example.py'])

    assert report == expected_report


def test_hide_garbage():
    """ Not a string or iterable collection of strings, ignore it. """
    code = """\
from space_tracer import traced


@traced(hide=42)
def foo(n):
    s = 'x'
    for i in range(n):
        s += 'y'
    return s

print(foo(3))
"""
    expected_report = """\
    @traced(hide=42)       |
    def foo(n):            | n = 3
        s = 'x'            | s = 'x'
        for i in range(n): | i = 0    | i = 1     | i = 2
            s += 'y'       | s = 'xy' | s = 'xyy' | s = 'xyyy'
        return s           | return 'xyyy'"""

    with replace_input(code):
        report = TraceRunner().trace_command([
            'space_tracer',
            '--source_indent', '4',
            '--traced_file', 'example.py',
            'example.py'])

    assert report == expected_report


def test_hide_from_command_line():
    code = """\
from space_tracer import traced


def foo(n):
    s = 'x'
    for i in range(n):
        s += 'y'
    return s

print(foo(3))
"""
    expected_report = """\
    def foo(n):            |
        s = 'x'            | s = 'x'
        for i in range(n): |          |           |
            s += 'y'       | s = 'xy' | s = 'xyy' | s = 'xyyy'
        return s           | return 'xyyy'"""

    with replace_input(code):
        report = TraceRunner().trace_command([
            'space_tracer',
            '--source_indent', '4',
            '--traced', 'foo',
            '--hide', 'n', 'i',
            '--traced_file', 'example.py',
            'example.py'])

    assert report == expected_report


def test_line_numbers():
    code = """\
def foo(n):
    s = 'x'
    for i in range(n):
        s += 'y'
    return s

print(foo(3))
"""
    expected_report = """\
    for i in range(n): | i = 0    | i = 1     | i = 2
        s += 'y'       | s = 'xy' | s = 'xyy' | s = 'xyyy'
    return s           | return 'xyyy'"""

    with replace_input(code):
        report = TraceRunner().trace_command([
            'space_tracer',
            '--start_line', '3',
            '--end_line', '5',
            '--traced_file', 'example.py',
            'example.py'])

    assert report == expected_report


def test_start_line_only():
    code = """\
def foo(n):
    s = 'x'
    for i in range(n):
        s += 'y'
    return s

print(foo(3))
"""
    expected_report = """\
    for i in range(n): | i = 0    | i = 1     | i = 2
        s += 'y'       | s = 'xy' | s = 'xyy' | s = 'xyyy'
    return s           | return 'xyyy'
                       |
print(foo(3))          | print('xyyy')"""

    with replace_input(code):
        report = TraceRunner().trace_command([
            'space_tracer',
            '--start_line', '3',
            '--traced_file', 'example.py',
            'example.py'])

    assert report == expected_report


def test_end_line_only():
    code = """\
def foo(n):
    s = 'x'
    for i in range(n):
        s += 'y'
    return s

print(foo(3))
"""
    expected_report = """\
def foo(n):            | n = 3
    s = 'x'            | s = 'x'
    for i in range(n): | i = 0    | i = 1     | i = 2"""

    with replace_input(code):
        report = TraceRunner().trace_command([
            'space_tracer',
            '--end_line', '3',
            '--traced_file', 'example.py',
            'example.py'])

    assert report == expected_report


def test_show_line_numbers():
    code = """\
a = 1
b = 2
print(a + b)
"""
    expected_report = """\
2) b = 2        | b = 2
3) print(a + b) | print('3')"""

    with replace_input(code):
        report = TraceRunner().trace_command([
            'space_tracer',
            '--start_line', '2',
            '--end_line', '3',
            '--line_numbers',
            '--traced_file', 'example.py',
            'example.py'])

    assert report == expected_report


def test_show_line_numbers_pad_right():
    code = """\
a = 1
b = 2
c = 3
d = 4
e = 5
f = 6
g = 7
h = 8
i = 9
print(f*g)"""
    expected_report = """\
 9) i = 9      | i = 9
10) print(f*g) | print('42')"""

    with replace_input(code):
        report = TraceRunner().trace_command([
            'space_tracer',
            '--start_line', '9',
            '--end_line', '10',
            '--line_numbers',
            '--traced_file', 'example.py',
            'example.py'])

    assert report == expected_report



from space_tracer.traced_finder import TracedFinder


def test_find_function():
    source = """\
# some comment

def foo(n):
    return n + 1
"""
    traced = 'foo'
    expected_line_number = 3

    finder = TracedFinder(source, traced)

    node = finder.traced_node
    assert expected_line_number == node.lineno


def test_named_function():
    source = """\
# some comment

def foo(n):
    return n + 1

def bar(n):
    return n - 1
"""
    traced = 'foo'
    expected_line_number = 3

    finder = TracedFinder(source, traced)

    node = finder.traced_node
    assert expected_line_number == node.lineno


def test_class():
    source = """\
class Foo(object):
    pass

class Bar(object):
    pass
"""
    traced = 'Bar'
    expected_line_number = 4

    finder = TracedFinder(source, traced)

    node = finder.traced_node
    assert expected_line_number == node.lineno


def test_method():
    source = """\
class Foo(object):
    def baz():
        pass

class Bar(object):
    def baz():
        pass
"""
    traced = 'Bar.baz'
    expected_line_number = 6

    finder = TracedFinder(source, traced)

    node = finder.traced_node
    assert expected_line_number == node.lineno


def test_nested_method():
    source = """\
def foo():
    def baz():
        pass

def bar():
    def baz():
        pass
"""
    traced = 'bar.baz'
    expected_line_number = 6

    finder = TracedFinder(source, traced)

    node = finder.traced_node
    assert expected_line_number == node.lineno


def test_suffix_not_allowed():
    source = """\
def foo(n):
    return n + 1

def bar(n):
    return n - 1
"""
    traced = 'parent.child.bar'

    finder = TracedFinder(source, traced)

    node = finder.traced_node
    assert node is None



import sys

import pytest

from space_tracer.main import TraceRunner
from test_report_builder import trim_report


def test_assignment():
    code = """\
i = 1
"""
    expected_report = """\
i = 1"""

    report = TraceRunner().trace_code(code)

    assert report == expected_report


@pytest.mark.skipif(
    sys.version_info < (3, 6),
    reason='type hints not available before Python 3.6')
def test_annotated_assignment():
    code = """\
i: int = 1
"""
    expected_report = """\
i = 1 """

    report = TraceRunner().trace_code(code)

    assert trim_report(expected_report) == trim_report(report)


@pytest.mark.skipif(
    sys.version_info < (3, 6),
    reason='type hints not available before Python 3.6')
def test_annotated_variable():
    code = """\
i: int
for i in range(2):
    pass
"""
    expected_report = """\

i = 0 | i = 1
      |"""

    report = TraceRunner().trace_code(code)

    assert report == expected_report


def test_increment():
    code = """\
i = 1
i += 1
"""
    expected_report = """\
i = 1
i = 2 """

    report = TraceRunner().trace_code(code)

    assert trim_report(expected_report) == trim_report(report)


def test_mutable():
    code = """\
a = [1, 2, [3, 4]]
a[0] = 9
a[2][1] = 8
b = a
i, j = 2, 1
a[i][j] = 7
a[0:2] = list(reversed(a[0:2]))
b = a
d = -1
a[i:j:d] = [100]
b = a
"""
    expected_report = """\
a = [1, 2, [3, 4]]
a[0] = 9
a[2][1] = 8
b = [9, 2, [3, 8]]
(i, j) = (2, 1)
a[2][1] = 7
a[0:2] = [2, 9]
b = [2, 9, [3, 7]]
d = -1
a[2:1:-1] = [100]
b = [2, 9, 100]
"""

    report = TraceRunner().trace_code(code)

    assert trim_report(expected_report) == trim_report(report)


def test_mutable_increment():
    code = """\
a = [1, 2, [3, 4]]
a[0] += 9
a[2][1] += 8
b = a
i, j = 2, 1
a[i][j] += 7
b = a
"""
    expected_report = """\
a = [1, 2, [3, 4]]
a[0] = 10
a[2][1] = 12
b = [10, 2, [3, 12]]
(i, j) = (2, 1)
a[2][1] = 19
b = [10, 2, [3, 19]]
"""

    report = TraceRunner().trace_code(code)

    assert trim_report(expected_report) == trim_report(report)


def test_slice():
    code = """\
a = [1, 2, 3, 4, 5]
i, j = 1, 4
a[i:j] = [20, 30]
b = a
a[2:] = [300]
b = a
a[:2] *= 2
b = a
a[:2] += [21]
b = a
"""
    expected_report = """\
a = [1, 2, 3, 4, 5]
(i, j) = (1, 4)
a[1:4] = [20, 30]
b = [1, 20, 30, 5]
a[2:] = [300]
b = [1, 20, 300]
a[:2] *= 2
b = [1, 20, 1, 20, 300]
a[:2] += [21]
b = [1, 20, 21, 1, 20, 300] """

    report = TraceRunner().trace_code(code)

    assert trim_report(expected_report) == trim_report(report)


def test_slice_magic():
    """ All augmented assignments on slices, possible with mocks. """
    code = """\
from unittest.mock import MagicMock


class Foo(MagicMock):
    def __repr__(self):
        return 'Foo()'

foo = Foo()
foo[1] = 3
foo[1:10] = 3
foo[1:10:2] = 3
foo[...] = 3
foo[1, 2:3] = 3
foo[1:10] += 3
foo[1:10] -= 3
foo[1:10] *= 3
foo[1:10] /= 3
foo[1:10] //= 3
foo[1:10] %= 3
foo[1:10] **= 3
foo[1:10] >>= 3
foo[1:10] <<= 3
foo[1:10] &= 3
foo[1:10] ^= 3
foo[1:10] |= 3
"""
    expected_report = """\







foo = Foo()
foo[1] = 3
foo[1:10] = 3
foo[1:10:2] = 3
foo[...] = 3
foo[1, 2:3] = 3
foo[1:10] += 3
foo[1:10] -= 3
foo[1:10] *= 3
foo[1:10] /= 3
foo[1:10] //= 3
foo[1:10] %= 3
foo[1:10] **= 3
foo[1:10] >>= 3
foo[1:10] <<= 3
foo[1:10] &= 3
foo[1:10] ^= 3
foo[1:10] |= 3 """

    report = TraceRunner().trace_code(code)

    assert trim_report(expected_report) == trim_report(report)


@pytest.mark.skipif(sys.version_info < (3, 8),
                    reason='Walrus operator added in 3.8.')
def test_walrus():
    code = """\
a = (b := 2) + 3
"""
    expected_report = """\
b = 2 | a = 5"""

    report = TraceRunner().trace_code(code)

    assert report == expected_report


def test_starred():
    code = """\
a, *b = 'apple'
print(a, b)
"""
    expected_report = """\
(a, *b) = ('a', 'p', 'p', 'l', 'e')
print("a ['p', 'p', 'l', 'e']")"""

    report = TraceRunner().trace_code(code)

    assert report == expected_report



import unittest

from space_tracer.report_builder import ReportBuilder


def trim_report(report):
    lines = report.splitlines()
    trimmed_lines = [line.rstrip() for line in lines]
    return '\n'.join(trimmed_lines) + '\n'


class ReportTestCase(unittest.TestCase):
    def setUp(self):
        super(ReportTestCase, self).setUp()
        self.addTypeEqualityFunc(str, self.assertMultiLineEqual)

    @staticmethod
    def assertReportEqual(expected_report, report):
        assert trim_report(expected_report) == trim_report(report)


class ReportBuilderTest(ReportTestCase):
    def setUp(self):
        if not hasattr(self, 'assertRaisesRegex'):
            self.assertRaisesRegex = self.assertRaisesRegexp

    def test_assign(self):
        # SETUP
        expected_value = 5
        expected_report = """\

x = 5
"""

        # EXEC
        builder = ReportBuilder()
        value = builder.assign(name='x', value=expected_value, line_number=2)
        report = builder.report()

        # VERIFY
        self.assertReportEqual(expected_report, report)
        self.assertEqual(expected_value, value)

    def test_assign_object_without_repr(self):
        # SETUP
        expected_report = ''

        class ClassWithoutRepr(object):
            pass

        # EXEC
        builder = ReportBuilder()
        builder.assign(name='x', value=ClassWithoutRepr(), line_number=2)
        report = builder.report()

        # VERIFY
        self.assertReportEqual(expected_report, report)

    def test_assign_object_with_multiline_repr(self):
        # SETUP
        expected_report = 'm = MultilineClass(1, 2) '

        class MultilineClass(object):
            def __init__(self, x, y):
                self.x = x
                self.y = y

            def __repr__(self):
                return "MultilineClass(%r,\n%r)" % (self.x, self.y)

        # EXEC
        builder = ReportBuilder()
        builder.assign(name='m', value=MultilineClass(1, 2), line_number=1)
        report = builder.report()

        # VERIFY
        self.assertReportEqual(expected_report, report)

    def test_assign_object_with_repr_exception(self):
        # SETUP
        expected_report = ''

        class BadReprClass(object):
            def __repr__(self):
                raise NotImplementedError()

        # EXEC
        builder = ReportBuilder()
        builder.assign(name='m', value=BadReprClass(), line_number=1)
        report = builder.report()

        # VERIFY
        self.assertReportEqual(expected_report, report)

    def test_assign_with_indexes(self):
        # SETUP
        expected_report = "a[0][23] = b['x'] = 99 "

        # EXEC
        builder = ReportBuilder()
        builder.start_assignment()
        builder.set_assignment_value(99)
        builder.add_assignment_index(0)
        builder.add_assignment_index(23)
        builder.add_assignment_index('x')
        builder.report_assignment('a[{!r}][{!r}] = b[{!r}] = {}',
                                  line_number=1)
        builder.end_assignment()
        report = builder.report()

        # VERIFY
        self.assertReportEqual(expected_report, report)

    def test_assign_value_after_indexes(self):
        # SETUP
        expected_report = "a[0][23] = 99 "

        # EXEC
        builder = ReportBuilder()
        builder.start_assignment()
        builder.add_assignment_index(0)
        builder.add_assignment_index(23)
        builder.set_assignment_value(99)
        builder.report_assignment('a[{!r}][{!r}] = {}',
                                  line_number=1)
        builder.end_assignment()
        report = builder.report()

        # VERIFY
        self.assertReportEqual(expected_report, report)

    def test_assign_function(self):
        # SETUP
        expected_report = ''

        def f(n):
            return n+1

        # EXEC
        builder = ReportBuilder()
        builder.assign(name='f', value=f, line_number=2)
        report = builder.report()

        # VERIFY
        self.assertReportEqual(expected_report, report)

    def test_lambda(self):
        # SETUP
        expected_report = '(23 => 24)'
        min_line = max_line = 1

        # EXEC
        builder = ReportBuilder()
        y = list(map(lambda x: builder.report_lambda(min_line, max_line, x, x+1),
                     [23]))
        report = builder.report()

        # VERIFY
        self.assertReportEqual(expected_report, report)
        self.assertEqual([24], y)

    def test_lambda_multiple_params(self):
        # SETUP
        expected_report = "('a', 'b' => 'ab')"
        min_line = max_line = 1

        # EXEC
        builder = ReportBuilder()
        z = list(map(lambda x, y: builder.report_lambda(min_line, max_line, x, y, x+y),
                     ['a'],
                     ['b']))
        report = builder.report()

        # VERIFY
        self.assertReportEqual(expected_report, report)
        self.assertEqual(['ab'], z)

    def test_lambda_multiple_calls(self):
        # SETUP
        expected_report = "('a', 'b' => 'ab') | (1, 2 => 3)"
        min_line = max_line = 1

        # EXEC
        builder = ReportBuilder()
        list(map(lambda x, y: builder.report_lambda(min_line, max_line, x, y, x+y),
                 ['a', 1],
                 ['b', 2]))
        report = builder.report()

        # VERIFY
        self.assertReportEqual(expected_report, report)

    def test_message_limit(self):
        # SETUP
        expected_report = """\
a = 10
b = 20
"""

        # EXEC
        builder = ReportBuilder(message_limit=4)
        builder.assign('a', 10, 1)
        builder.assign('b', 20, 2)
        with self.assertRaisesRegex(RuntimeError, r'message limit exceeded'):
            builder.assign('c', 30, 3)
        report = builder.report()

        # VERIFY
        self.assertReportEqual(expected_report, report)

    def test_message_limit_with_empty_blocks(self):
        # SETUP
        expected_report = """\




"""

        # EXEC
        builder = ReportBuilder(message_limit=2)
        builder.start_frame(1, 5)
        builder.start_block(1, 3)
        builder.start_block(1, 3)
        report = builder.report()

        # VERIFY
        self.assertReportEqual(expected_report, report)

    def test_return(self):
        # SETUP
        expected_value = 'bob'
        expected_report = """\

return 'bob'
"""

        # EXEC
        builder = ReportBuilder()
        value = builder.return_value(value=expected_value, line_number=2)
        report = builder.report()

        # VERIFY
        self.assertReportEqual(expected_report, report)
        self.assertEqual(expected_value, value)

    def test_return_multiline(self):
        # SETUP
        class Multiline(object):
            def __repr__(self):
                return 'Multiline(\n)'
        expected_value = Multiline()
        expected_report = """\

return Multiline( )
"""

        # EXEC
        builder = ReportBuilder()
        value = builder.return_value(value=expected_value, line_number=2)
        report = builder.report()

        # VERIFY
        self.assertReportEqual(expected_report, report)
        self.assertEqual(expected_value, value)

    def test_call(self):
        # SETUP
        expected_report = """\

a = [1, 2]
"""
        expected_result = 23

        # EXEC
        builder = ReportBuilder()
        result = builder.record_call(names=['a'],
                                     displays_before=['[2, 1]'],
                                     result=expected_result,
                                     displays_after=['[1, 2]'],
                                     line_number=2)
        report = builder.report()

        # VERIFY
        self.assertReportEqual(expected_report, report)
        self.assertEqual(expected_result, result)

    def test_call_multiple_changes(self):
        # SETUP
        expected_report = """\

a = [1, 2] b = 6
"""
        expected_result = 23

        # EXEC
        builder = ReportBuilder()
        result = builder.record_call(names=['a', 'b'],
                                     displays_before=['[2, 1]', '5'],
                                     result=expected_result,
                                     displays_after=['[1, 2]', '6'],
                                     line_number=2)
        report = builder.report()

        # VERIFY
        self.assertReportEqual(expected_report, report)
        self.assertEqual(expected_result, result)

    def test_call_no_change(self):
        # SETUP
        expected_report = ""
        expected_result = 23

        # EXEC
        builder = ReportBuilder()
        result = builder.record_call(names=['a'],
                                     displays_before=['[1, 2]'],
                                     result=expected_result,
                                     displays_after=['[1, 2]'],
                                     line_number=2)
        report = builder.report()

        # VERIFY
        self.assertReportEqual(expected_report, report)
        self.assertEqual(expected_result, result)

    def test_delete_item(self):
        # SETUP
        expected_report = "d = {'a': 1}"
        d = {'a': 1, 'b': 2}
        expected_d = {'a': 1}

        # EXEC
        builder = ReportBuilder()
        del builder.record_delete(name='d', target=d, line_number=1)['b']
        report = builder.report()

        # VERIFY
        self.assertReportEqual(expected_report, report)
        self.assertEqual(expected_d, d)

    def test_multiple_lines(self):
        # SETUP
        expected_report = """\
a = 3

b = 'xyz' """

        # EXEC
        builder = ReportBuilder()
        builder.assign(name='a', value=3, line_number=1)
        builder.assign(name='b', value="xyz", line_number=3)
        report = builder.report()

        # VERIFY
        self.assertReportEqual(expected_report, report)

    def test_multiple_messages(self):
        # SETUP
        expected_report = """\
x = 1 | y = 2 """

        # EXEC
        builder = ReportBuilder()
        builder.assign(name='x', value=1, line_number=1)
        builder.assign(name='y', value=2, line_number=1)
        report = builder.report()

        # VERIFY
        self.assertReportEqual(expected_report, report)

    def test_multiple_visits(self):
        # SETUP
        expected_report = """\
x = 'first'    | x = 'second'
y = 'continue' | """

        # EXEC
        builder = ReportBuilder()
        builder.start_block(1, 2)
        builder.assign(name='x', value='first', line_number=1)
        builder.assign(name='y', value='continue', line_number=2)
        builder.start_block(1, 2)
        builder.assign(name='x', value='second', line_number=1)
        report = builder.report()

        # VERIFY
        self.assertReportEqual(expected_report, report)

    def test_nested_blocks(self):
        # SETUP
        expected_report = """\
x = 'first' | x = 'second'
            | y = 1 | y = 2 """

        # EXEC
        builder = ReportBuilder()
        builder.start_block(1, 2)
        builder.assign(name='x', value='first', line_number=1)
        builder.start_block(1, 2)
        builder.assign(name='x', value='second', line_number=1)
        builder.start_block(2, 2)
        builder.assign(name='y', value=1, line_number=2)
        builder.start_block(2, 2)
        builder.assign(name='y', value=2, line_number=2)
        report = builder.report()

        # VERIFY
        self.assertReportEqual(expected_report, report)

    def test_separate_scopes(self):
        # SETUP
        expected_report = """\
x = 'child' | x = 'child again'

y = 'main' """

        # EXEC
        builder = ReportBuilder()
        builder.assign(name='y', value='main', line_number=3)
        builder.start_block(1, 1)
        builder.assign(name='x', value='child', line_number=1)
        builder.start_block(1, 1)
        builder.assign(name='x', value='child again', line_number=1)
        report = builder.report()

        # VERIFY
        self.assertReportEqual(expected_report, report)

    def test_frames(self):
        # SETUP
        expected_report = """\
i = 1 | i = 2
n = 2 | """

        # EXEC
        builder = ReportBuilder()
        frame1 = builder.start_frame(1, 2)
        frame1.assign(name='i', value=1, line_number=1)
        frame2 = builder.start_frame(1, 2)
        frame2.assign(name='i', value=2, line_number=1)
        frame1.assign(name='n', value=2, line_number=2)
        report = builder.report()

        # VERIFY
        self.assertReportEqual(expected_report, report)

    def test_frames_with_extra_message(self):
        # SETUP
        expected_report = """\
i = 1 | i = 2
      | extra message"""

        # EXEC
        builder = ReportBuilder()
        frame1 = builder.start_frame(1, 2)
        frame1.assign(name='i', value=1, line_number=1)
        frame2 = builder.start_frame(1, 2)
        frame2.assign(name='i', value=2, line_number=1)
        builder.add_extra_message('extra message', 2)
        report = builder.report()

        # VERIFY
        self.assertReportEqual(expected_report, report)

    def test_decorated_frame(self):
        # SETUP
        expected_report = """\
i = 1 | i = 2
n = 2 | """

        # EXEC
        builder = ReportBuilder()
        builder.assign(name='i', value=0, line_number=2)
        ReportBuilder.is_tracing_next_block = True
        frame1 = builder.start_frame(3, 4)
        frame1.assign(name='i', value=1, line_number=3)
        ReportBuilder.is_tracing_next_block = True
        frame2 = builder.start_frame(3, 4)
        frame2.assign(name='i', value=2, line_number=3)
        frame1.assign(name='n', value=2, line_number=4)
        builder.assign(name='i', value=20, line_number=5)
        report = builder.report()

        # VERIFY
        self.assertReportEqual(expected_report, report)

    def test_trace_block(self):
        # SETUP
        expected_report = """\
i = 1 | i = 2
n = 2 | """

        # EXEC
        builder = ReportBuilder()
        builder.trace_block(3, 4)
        builder.assign(name='i', value=0, line_number=2)
        frame1 = builder.start_frame(3, 4)
        frame1.assign(name='i', value=1, line_number=3)
        frame2 = builder.start_frame(3, 4)
        frame2.assign(name='i', value=2, line_number=3)
        frame1.assign(name='n', value=2, line_number=4)
        builder.assign(name='i', value=20, line_number=5)
        report = builder.report()

        # VERIFY
        self.assertReportEqual(expected_report, report)

    def test_exception(self):
        # SETUP
        source = """\
try:
    raise RuntimeError('foo')
except:
    builder.exception()"""
        expected_report = """\

RuntimeError: foo
"""

        # EXEC
        builder = ReportBuilder()
        environment = dict(builder=builder)
        exec(source, environment, environment)
        report = builder.report()

        # VERIFY
        self.assertReportEqual(expected_report, report)

    def test_exception_multiline_message(self):
        # SETUP
        source = """\
try:
    raise RuntimeError('a\\nb')
except:
    builder.exception()"""
        expected_report = """\

RuntimeError: a b
"""

        # EXEC
        builder = ReportBuilder()
        environment = dict(builder=builder)
        exec(source, environment, environment)
        report = builder.report()

        # VERIFY
        self.assertReportEqual(expected_report, report)

    def test_message_count(self):
        # SETUP

        # EXEC
        builder = ReportBuilder()
        builder.add_message('a', 1)
        builder.add_message('b', 2)

        # VERIFY
        self.assertEqual(2, builder.message_count)

    def test_count_all_messages(self):
        # SETUP

        # EXEC
        builder = ReportBuilder()
        frame = builder.start_frame(1, 2)
        frame.add_message('a', 1)
        frame.add_message('b', 2)

        # VERIFY
        self.assertEqual(0, builder.message_count)
        self.assertEqual(2, builder.count_all_messages())

    def test_output(self):
        # SETUP
        expected_report = """\
print('a')
"""

        # EXEC
        builder = ReportBuilder()
        builder.add_output('a\n', 1)

        # VERIFY
        self.assertReportEqual(expected_report, builder.report())

    def test_multiple_outputs(self):
        # SETUP
        expected_report = """\
print('a\\nb')
"""

        # EXEC
        builder = ReportBuilder()
        builder.add_output('a\n', 1)
        builder.add_output('b\n', 1)

        # VERIFY
        self.assertReportEqual(expected_report, builder.report())

    def test_multiple_outputs_different_lines(self):
        # SETUP
        expected_report = """\
print('a')

print('b')
"""

        # EXEC
        builder = ReportBuilder()
        builder.add_output('a\n', 1)
        builder.add_output('b\n', 3)

        # VERIFY
        self.assertReportEqual(expected_report, builder.report())

    def test_output_without_newline(self):
        # SETUP
        expected_report = """\
sys.stdout.write('a')
"""

        # EXEC
        builder = ReportBuilder()
        builder.add_output('a', 1)

        # VERIFY
        self.assertReportEqual(expected_report, builder.report())

    def test_output_stderr(self):
        # SETUP
        expected_report = """\
sys.stderr.write('a\\n')
"""

        # EXEC
        builder = ReportBuilder()
        builder.add_output('a\n', 1, target_name='sys.stderr')

        # VERIFY
        self.assertReportEqual(expected_report, builder.report())

    def test_format_stderr(self):
        # SETUP
        expected_report = "sys.stderr.write('abc') | x = None"

        # EXEC
        builder = ReportBuilder()
        builder.add_output('abc', 1, target_name='sys.stderr')
        builder.assign(name='x', value=None, line_number=1)

        # VERIFY
        self.assertReportEqual(expected_report, builder.report())


if __name__ == '__main__':
    unittest.main()



from space_tracer.main import TraceRunner, replace_input
from test_report_builder import trim_report


def test_print():
    code = """\
s = 'x'
print(s)
"""
    expected_report_python = """\
s = 'x'
print('x')
"""
    expected_report = expected_report_python
    tracer = TraceRunner()

    report = tracer.trace_code(code)

    assert expected_report == trim_report(report)


def test_print_in_function():
    code = """\
from __future__ import print_function
def main():
    s = 'Hello, World!'
    print(s)

main()
"""
    expected_report = """\


s = 'Hello, World!'
print('Hello, World!')

"""
    tracer = TraceRunner()

    report = tracer.trace_code(code)

    assert expected_report == trim_report(report)


def test_print_in_loop():
    code = """\
from __future__ import print_function
for i in range(3):
    print(i)
"""
    expected_report = """\

i = 0      | i = 1      | i = 2
print('0') | print('1') | print('2')
"""
    tracer = TraceRunner()

    report = tracer.trace_code(code)

    assert expected_report == trim_report(report)


def test_stdout(capsys):
    code = """\
import sys
s = 'x'
sys.stdout.write(s)
"""
    expected_report = """\

s = 'x'
sys.stdout.write('x')
"""
    tracer = TraceRunner()

    report = tracer.trace_code(code)

    assert expected_report == trim_report(report)
    std = capsys.readouterr()
    assert "" == std.out


def test_stderr(capsys):
    code = """\
import sys
s = 'x'
sys.stderr.write(s)
"""
    expected_report = """\

s = 'x'
sys.stderr.write('x')
"""
    tracer = TraceRunner()

    report = tracer.trace_code(code)

    assert expected_report == trim_report(report)
    std = capsys.readouterr()
    assert "" == std.err


def test_string_io():
    code = """\
from __future__ import unicode_literals
from io import StringIO

f = StringIO()
f.write('x')
"""
    expected_report_python = """\




f.write('x')
"""
    expected_report = expected_report_python
    tracer = TraceRunner()

    report = tracer.trace_code(code)

    assert expected_report == trim_report(report)


def test_print_with_sep():
    code = """\
from __future__ import print_function

p = 'Bob'
n = 23
s = '--'
print(p, n, sep=s)
"""
    expected_report = """\


p = 'Bob'
n = 23
s = '--'
print('Bob--23')
"""

    report = TraceRunner().trace_code(code)

    assert expected_report == trim_report(report)


def test_print_with_star():
    code = """\
from __future__ import print_function

args = ['Bob', 23]
print(*args)
"""
    expected_report = """\


args = ['Bob', 23]
print('Bob 23')
"""

    report = TraceRunner().trace_code(code)

    assert expected_report == trim_report(report)


def test_report_default():
    code = """\
print('Hello,')
print(6*7)
"""
    expected_report = """\
print('Hello,')
print('42')"""

    with replace_input(code):
        report = TraceRunner().trace_command([
            'space_tracer',
            '--report', '-',
            '--source_width', '0',
            '--traced_file', 'example.py',
            'example.py'])

    assert report == expected_report


def test_report_redirect(tmpdir):
    code = """\
print('Hello,')
print(6*7)
"""
    expected_report = """\
print('Hello,')
print('42')"""

    report_path = tmpdir.join('report.txt')
    with replace_input(code):
        stdout_report = TraceRunner().trace_command([
            'space_tracer',
            '--report', str(report_path),
            '--source_width', '0',
            '--traced_file', 'example.py',
            'example.py'])
    report = report_path.read_text('utf8')

    assert stdout_report == ''
    assert report == expected_report


def test_report_redirect_to_stderr(capsys):
    code = """\
print('Hello,')
print(6*7)
"""
    expected_report = """\
print('Hello,')
print('42')"""

    with replace_input(code):
        stdout_report = TraceRunner().trace_command([
            'space_tracer',
            '--report', '!',
            '--source_width', '0',
            '--traced_file', 'example.py',
            'example.py'])
    report = capsys.readouterr().err

    assert stdout_report == ''
    assert report == expected_report


def test_stdout_redirect(tmpdir):
    code = """\
print('Hello,')
print(6*7)
"""
    expected_report = """\
print('Hello,')
print('42')"""
    expected_redirect = """\
Hello,
42
"""

    stdout_path = tmpdir.join('stdout.txt')
    with replace_input(code):
        report = TraceRunner().trace_command([
            'space_tracer',
            '--stdout', str(stdout_path),
            '--source_width', '0',
            '--traced_file', 'example.py',
            'example.py'])
    stdout_redirect = stdout_path.read_text('utf8')

    assert report == expected_report
    assert stdout_redirect == expected_redirect


def test_stderr_redirect(tmpdir):
    code = """\
import sys
print('Hello,', file=sys.stderr)
print(6*7)
"""
    expected_report = """\

sys.stderr.write('Hello,\\n')
print('42')"""
    expected_redirect = """\
Hello,
"""

    stderr_path = tmpdir.join('stderr.txt')
    with replace_input(code):
        report = TraceRunner().trace_command([
            'space_tracer',
            '--stderr', str(stderr_path),
            '--source_width', '0',
            '--traced_file', 'example.py',
            'example.py'])
    stderr_redirect = stderr_path.read_text('utf8')

    assert report == expected_report
    assert stderr_redirect == expected_redirect


def test_redirect_both(tmpdir):
    code = """\
import sys
print('Hello,', file=sys.stderr)
print(6*7)
"""
    expected_report = """\

sys.stderr.write('Hello,\\n')
print('42')"""
    expected_redirect = """\
Hello,
42
"""

    out_path = tmpdir.join('out.txt')
    with replace_input(code):
        report = TraceRunner().trace_command([
            'space_tracer',
            '--stdout', str(out_path),
            '--stderr', str(out_path),
            '--source_width', '0',
            '--traced_file', 'example.py',
            'example.py'])
    out_redirect = out_path.read_text('utf8')

    assert report == expected_report
    assert out_redirect == expected_redirect


def test_string_io_field():
    code = """\
from __future__ import unicode_literals
from io import StringIO


class Foo:
    def __init__(self, f):
        self.f = f

    def greet(self, name):
        self.f.write('Hello, ' + name + '.')

greetings = StringIO()
foo = Foo(greetings)
foo.greet('Alice')
"""
    expected_report_python = """\








name = 'Alice'
self.f.write('Hello, Alice.')



"""
    expected_report = expected_report_python

    report = TraceRunner().trace_code(code)

    assert expected_report == trim_report(report)


def test_no_input():
    code = """\
s = input()
"""
    expected_report = """\
EOFError: EOF when reading a line
"""

    report = TraceRunner().trace_code(code)

    assert expected_report == trim_report(report)


def test_input(tmpdir):
    input_text = """\
first line
second line
"""
    code = "s = input()"
    expected_report = """\
s = 'first line'
"""
    stdin_path = str(tmpdir.join('stdin.txt'))
    with open(stdin_path, 'w') as f:
        f.write(input_text)

    with replace_input(code):
        report = TraceRunner().trace_command([
            'space_tracer',
            '--stdin', stdin_path,
            '--source_width', '0',
            '--traced_file', 'example.py',
            'example.py'])

    assert expected_report == trim_report(report)


def test_input_stdin(tmpdir):
    input_text = """\
first line
second line
"""
    code = "s = input()"
    expected_report = """\
s = 'first line'
"""
    code_path = str(tmpdir.join('example.py'))
    with open(code_path, 'w') as f:
        f.write(code)

    with replace_input(input_text):
        report = TraceRunner().trace_command([
            'space_tracer',
            '--stdin', '-',
            '--source_width', '0',
            code_path])

    assert expected_report == trim_report(report)


def test_default_input(tmpdir):
    code = "s = input()"
    expected_report = """\
EOFError: EOF when reading a line
"""

    with replace_input(code):
        report = TraceRunner().trace_command([
            'space_tracer',
            '--source_width', '0',
            '--traced_file', 'example.py',
            'example.py'])

    assert expected_report == trim_report(report)


def test_prompt(tmpdir):
    input_text = """\
first line
second line
"""
    code = "s = input('What comes first?')"
    expected_report = """\
sys.stdout.write('What comes first?') | s = 'first line'
"""
    stdin_path = str(tmpdir.join('stdin.txt'))
    with open(stdin_path, 'w') as f:
        f.write(input_text)

    with replace_input(code):
        report = TraceRunner().trace_command([
            'space_tracer',
            '--stdin', stdin_path,
            '--source_width', '0',
            '--traced_file', 'example.py',
            'example.py'])

    assert expected_report == trim_report(report)



from space_tracer.main import TraceRunner


def test_loop():
    code = """\
i = 1
for j in range(3):
    i += j
"""
    expected_report = """\
i = 1
j = 0 | j = 1 | j = 2
i = 1 | i = 2 | i = 4"""

    report = TraceRunner().trace_code(code)

    assert report == expected_report


def test_loop_target_list():
    code = """\
for a,b in [(1,2)]:
    c = a + b
"""
    expected_report = """\
a = 1 | b = 2
c = 3"""
    tracer = TraceRunner()

    report = tracer.trace_code(code)

    assert report == expected_report


def test_loop_starred_target_list():
    code = """\
words = ['foo', 'bar']
for (head, *tail) in words:
    print(head, tail)
"""
    expected_report = """\
words = ['foo', 'bar']
head = 'f' | tail = ['o', 'o'] | head = 'b' | tail = ['a', 'r']
print("f ['o', 'o']")          | print("b ['a', 'r']")"""
    tracer = TraceRunner()

    report = tracer.trace_code(code)

    assert report == expected_report


def test_loop_target_list_attribute():
    code = """\
class Foo:
    def do_foo(self, x):
        for self.i in range(x):
            print(self.i)

foo = Foo()
foo.do_foo(3)
"""
    expected_report = """\

x = 3
self.i = 0 | self.i = 1 | self.i = 2
print('0') | print('1') | print('2')


"""
    tracer = TraceRunner()

    report = tracer.trace_code(code)

    assert report == expected_report


def test_loop_target_list_attribute_complex():
    code = """\
class Foo:
    def do_foo(self, x):
        self.state = [None]
        for self.state[0] in range(x):
            print(self.state)

foo = Foo()
foo.do_foo(3)
"""
    expected_report = """\

x = 3
self.state = [None]
             |              |
print('[0]') | print('[1]') | print('[2]')


"""
    tracer = TraceRunner()

    report = tracer.trace_code(code)

    assert report == expected_report


def test_nested_loop():
    code = """\
n = 0
for i in range(2):
    n += i
    for j in range(3):
        n += j
"""
    expected_report = """\
n = 0
i = 0                 | i = 1
n = 0                 | n = 4
j = 0 | j = 1 | j = 2 | j = 0 | j = 1 | j = 2
n = 0 | n = 1 | n = 3 | n = 4 | n = 5 | n = 7"""
    report = TraceRunner().trace_code(code)

    assert report == expected_report


def test_for_else():
    code = """\
i = 1
for j in range(3):
    i += j
else:
    i *= 10
"""
    expected_report = """\
i = 1
j = 0 | j = 1 | j = 2
i = 1 | i = 2 | i = 4

i = 40"""
    report = TraceRunner().trace_code(code)

    assert report == expected_report


def test_while_else():
    code = """\
i = 0
while i < 2:
    i += 1
else:
    i *= 10
"""
    expected_report = """\
i = 0
      |
i = 1 | i = 2

i = 20"""
    report = TraceRunner().trace_code(code)

    assert report == expected_report


def test_loop_conditional():
    code = """\
for i in range(3):
    if i == 1:
        c = 5
c = 2
"""
    expected_report = """\
i = 0 | i = 1 | i = 2
      |       |
      | c = 5 |
c = 2"""
    tracer = TraceRunner()

    report = tracer.trace_code(code)

    assert report == expected_report


def test_infinite_loop_by_count():
    code = """\
n = 0
while True:
    n += 1
"""
    expected_report = """\
n = 0
      |       |
n = 1 | n = 2 | RuntimeError: live coding message limit exceeded"""
    tracer = TraceRunner()
    tracer.message_limit = 8

    report = tracer.trace_code(code)

    assert report == expected_report


def test_infinite_loop_by_width():
    code = """\
n = 0
while True:
    n += 1
"""
    expected_report = """\
n = 0
      |       |
n = 1 | n = 2 | RuntimeError: live coding message limit exceeded"""
    tracer = TraceRunner()
    tracer.max_width = 20

    report = tracer.trace_code(code)

    assert report == expected_report


def test_infinite_loop_pass():
    code = """\
while True:
    pass
"""
    expected_report = """\
RuntimeError: live coding message limit exceeded
"""
    tracer = TraceRunner()
    tracer.message_limit = 3

    report = tracer.trace_code(code)

    assert report == expected_report


def test_infinite_loop_pass_in_function():
    code = """\
def foo():
    while True:
        pass

foo()
"""
    expected_report = """\

RuntimeError: live coding message limit exceeded


RuntimeError: live coding message limit exceeded"""
    tracer = TraceRunner()
    tracer.message_limit = 3

    report = tracer.trace_code(code)

    assert report == expected_report



import sys

try:
    from .main import main, web_main, traced
    from .about import __title__, __version__, __url__
except (ImportError, SyntaxError):
    if sys.version_info < (3,):
        print('Python 2 is no longer supported.')
        exit(1)



import unittest
from space_tracer.canvas import Canvas


class CanvasTest(unittest.TestCase):
    def setUp(self):
        if not hasattr(self, 'assertRaisesRegex'):
            self.assertRaisesRegex = self.assertRaisesRegexp

    def test_create_line(self):
        # SETUP
        expected_report = """\
create_line
    1
    2
    100
    200"""
        
        # EXEC
        canvas = Canvas()
        canvas.create_line(1, 2, 100, 200)
        report = canvas.report
        
        # VERIFY
        self.assertEqual(expected_report.splitlines(), report)

    def test_multiple_calls(self):
        # SETUP
        expected_report = """\
create_line
    1
    2
    100
    200
create_rectangle
    5
    10
    500
    1000"""
        
        # EXEC
        canvas = Canvas()
        canvas.create_line(1, 2, 100, 200)
        canvas.create_rectangle(5, 10, 500, 1000)
        report = canvas.report
        
        # VERIFY
        self.assertEqual(expected_report.splitlines(), report)
        
    def test_unknown_method(self):
        # EXEC
        canvas = Canvas()
        with self.assertRaisesRegex(
                AttributeError, 
                "(Canvas instance|'Canvas' object) has no attribute 'create_wirple'"):
            canvas.create_wirple(1, 'floop')

    def test_bounds(self):
        # SETUP
        expected_width = 800
        expected_height = 600
        # EXEC
        canvas = Canvas(expected_width, expected_height)
        width = canvas.cget('width')
        height = canvas.cget('height')
        
        # VERIFY
        self.assertEqual(expected_width, width)
        self.assertEqual(expected_height, height)

    def test_create_text(self):
        # SETUP
        expected_report = """\
create_text
    100
    200
    anchor='s'
    font=('Arial', 8, 'normal')
    text='foo'"""
        
        # EXEC
        canvas = Canvas()
        canvas.create_text(100, 
                           200, 
                           text='foo', 
                           font=('Arial', 8, 'normal'),
                           anchor='s')
        report = canvas.report
        
        # VERIFY
        self.assertEqual(expected_report.splitlines(), report)

if __name__ == '__main__':
    unittest.main()
